"""
Barter system

Evennia contribution - Griatch 2012


This implements a full barter system - a way for players to safely
trade items between each other using code rather than simple free-form
talking.  The advantage of this is increased buy/sell safety but it
also streamlines the process and makes it faster when doing many
transactions (since goods are automatically exchanged once both
agree).

This system is primarily intended for a barter economy, but can easily
be used in a monetary economy as well -- just let the "goods" on one
side be coin objects (this is more flexible than a simple "buy"
command since you can mix coins and goods in your trade).

In this module, a "barter" is generally referred to as a "trade".


- Trade example

A trade (barter) action works like this: A and B are the parties.

1) opening a trade

A: trade B: Hi, I have a nice extra sword. You wanna trade?
B sees: A says: "Hi, I have a nice extra sword. You wanna trade?"
   A wants to trade with you. Enter 'trade A <emote>' to accept.
B: trade A: Hm, I could use a good sword ...
A sees: B says: "Hm, I could use a good sword ...
   B accepts the trade. Use 'trade help' for aid.
B sees: You are now trading with A. Use 'trade help' for aid.

2) negotiating

A: offer sword: This is a nice sword. I would need some rations in trade.
B sees: A says: "This is a nice sword. I would need some rations in trade."
   [A offers Sword of might.]
B evaluate sword
B sees: <Sword's description and possibly stats>
B: offer ration: This is a prime ration.
A sees: B says: "This is a prime ration."
  [B offers iron ration]
A: say Hey, this is a nice sword, I need something more for it.
B sees: A says: "Hey this is a nice sword, I need something more for it."
B: offer sword,apple: Alright. I will also include a magic apple. That's my last offer.
A sees: B says: "Alright, I will also include a magic apple. That's my last offer."
  [B offers iron ration and magic apple]
A accept: You are killing me here, but alright.
B sees: A says: "You are killing me here, but alright."
  [A accepts your offer. You must now also accept.]
B accept: Good, nice making business with you.
  You accept the deal. Deal is made and goods changed hands.
A sees: B says: "Good, nice making business with you."
  B accepts the deal. Deal is made and goods changed hands.

At this point the trading system is exited and the negotiated items
are automatically exchanged between the parties. In this example B was
the only one changing their offer, but also A could have changed their
offer until the two parties found something they could agree on. The
emotes are optional but useful for RP-heavy worlds.

- Technical info

The trade is implemented by use of a TradeHandler. This object is a
common place for storing the current status of negotiations. It is
created on the object initiating the trade, and also stored on the
other party once that party agrees to trade. The trade request times
out after a certain time - this is handled by a Script. Once trade
starts, the CmdsetTrade cmdset is initiated on both parties along with
the commands relevant for the trading.

- Ideas for NPC bartering:

This module is primarily intended for trade between two players. But
it can also in principle be used for a player negotiating with an
AI-controlled NPC. If the NPC uses normal commands they can use it
directly -- but more efficient is to have the NPC object send its
replies directly through the tradehandler to the player. One may want
to add some functionality to the decline command, so players can
decline specific objects in the NPC offer (decline <object>) and allow
the AI to maybe offer something else and make it into a proper
barter.  Along with an AI that "needs" things or has some sort of
personality in the trading, this can make bartering with NPCs at least
moderately more interesting than just plain 'buy'.

- Installation:

Just import the CmdTrade command into (for example) the default
cmdset. This will make the trade (or barter) command available
in-game.

"""

from evennia.commands.cmdset import CmdSet
from evennia.commands.command import Command
from evennia.scripts.scripts import DefaultScript

TRADE_TIMEOUT = 60  # timeout for B to accept trade


class TradeTimeout(DefaultScript):
    """
    This times out the trade request, in case player B did not reply in time.
    """

    def at_script_creation(self):
        """
        Called when script is first created
        """
        self.key = "trade_request_timeout"
        self.desc = "times out trade requests"
        self.interval = TRADE_TIMEOUT
        self.start_delay = True
        self.repeats = 1
        self.persistent = False

    def at_repeat(self):
        """
        called once
        """
        if self.ndb.tradeevent:
            self.obj.ndb.tradeevent.finish(force=True)
        self.obj.msg("Trade request timed out.")

    def is_valid(self):
        """
        Only valid if the trade has not yet started
        """
        return self.obj.ndb.tradeevent and not self.obj.ndb.tradeevent.trade_started


class TradeHandler(object):
    """
    Objects of this class handles the ongoing trade, notably storing the current
    offers from each side and wether both have accepted or not.
    """

    def __init__(self, part_a, part_b):
        """
        Initializes the trade. This is called when part A tries to
        initiate a trade with part B. The trade will not start until
        part B repeats this command (B will then call the self.join()
        command)

        Args:
            part_a (object): The party trying to start barter.
            part_b (object): The party asked to barter.

        Notes:
            We also store the back-reference from the respective party
            to this object.

        """
        # parties
        self.part_a = part_a
        self.part_b = part_b

        self.part_a.cmdset.add(CmdsetTrade())
        self.trade_started = False
        self.part_a.ndb.tradehandler = self
        # trade variables
        self.part_a_offers = []
        self.part_b_offers = []
        self.part_a_accepted = False
        self.part_b_accepted = False

    def msg_other(self, sender, string):
        """
        Relay a message to the *other* party without needing to know
        which party that is. This allows the calling command to not
        have to worry about which party they are in the handler.

        Args:
            sender (object): One of A or B. The method will figure
                out the *other* party to send to.
            string (str): Text to send.
        """
        if self.part_a == sender:
            self.part_b.msg(string)
        elif self.part_b == sender:
            self.part_a.msg(string)
        else:
            # no match, relay to oneself
            sender.msg(string) if sender else self.part_a.msg(string)

    def get_other(self, party):
        """
        Returns the other party of the trade

        Args:
            party (object): One of the parties of the negotiation

        Returns:
            party_other (object): The other party, not the first party.

        """
        if self.part_a == party:
            return self.part_b
        if self.part_b == party:
            return self.part_a
        return None

    def join(self, part_b):
        """
        This is used once B decides to join the trade

        Args:
            part_b (object): The party accepting the barter.

        """
        if self.part_b == part_b:
            self.part_b.ndb.tradehandler = self
            self.part_b.cmdset.add(CmdsetTrade())
            self.trade_started = True
            return True
        return False

    def unjoin(self, part_b):
        """
        This is used if B decides not to join the trade.

        Args:
            part_b (object): The party leaving the barter.

        """
        if self.part_b == part_b:
            self.finish(force=True)
            return True
        return False

    def offer(self, party, *args):
        """
        Change the current standing offer. We leave it up to the
        command to do the actual checks that the offer consists
        of real, valid, objects.

        Args:
            party (object): Who is making the offer
            args (objects or str): Offerings.

        """
        if self.trade_started:
            # reset accept statements whenever an offer changes
            self.part_a_accepted = False
            self.part_b_accepted = False
            if party == self.part_a:
                self.part_a_offers = list(args)
            elif party == self.part_b:
                self.part_b_offers = list(args)
            else:
                raise ValueError

    def list(self):
        """
        List current offers.

        Returns:
            offers (tuple): A tuple with two lists, (A_offers, B_offers).

        """
        return self.part_a_offers, self.part_b_offers

    def search(self, offername):
        """
        Search current offers.

        Args:
            offername (str or int): Object to search for, or its index in
                the list of offered items.

        Returns:
            offer (object): An object on offer, based on the search criterion.

        """
        all_offers = self.part_a_offers + self.part_b_offers
        if isinstance(offername, int):
            # an index to return
            if 0 <= offername < len(all_offers):
                return all_offers[offername]

        all_keys = [offer.key for offer in all_offers]
        try:
            imatch = all_keys.index(offername)
            return all_offers[imatch]
        except ValueError:
            for offer in all_offers:
                if offer.aliases.get(offername):
                    return offer
        return None

    def accept(self, party):
        """
        Accept the current offer.

        Args:
            party (object): The party accepting the deal.

        Returns:
            result (object): `True` if this closes the deal, `False`
                otherwise

        Notes:
            This will only close the deal if both parties have
            accepted independently. This is done by calling the
            `finish()` method.

        """
        if self.trade_started:
            if party == self.part_a:
                self.part_a_accepted = True
            elif party == self.part_b:
                self.part_b_accepted = True
            else:
                raise ValueError
            return self.finish()  # try to close the deal
        return False

    def decline(self, party):
        """
        Decline the offer (or change one's mind).

        Args:
            party (object): Party declining the deal.

        Returns:
            did_decline (bool): `True` if there was really an
                `accepted` status to change, `False` otherwise.

        Notes:
            If previously having used the `accept` command, this
            function will only work as long as the other party has not
            yet accepted.

        """
        if self.trade_started:
            if party == self.part_a:
                if self.part_a_accepted:
                    self.part_a_accepted = False
                    return True
                return False
            elif party == self.part_b:
                if self.part_b_accepted:
                    self.part_b_accepted = False
                    return True
                return False
            else:
                raise ValueError
        return False

    def finish(self, force=False):
        """
        Conclude trade - move all offers and clean up

        Args:
            force (bool, optional): Force cleanup regardless of if the
                trade was accepted or not (if not, no goods will change
                hands but trading will stop anyway)
        Returns:
            result (bool): If the finish was successful.

        """
        fin = False
        if self.trade_started and self.part_a_accepted and self.part_b_accepted:
            # both accepted - move objects before cleanup
            for obj in self.part_a_offers:
                obj.location = self.part_b
            for obj in self.part_b_offers:
                obj.location = self.part_a
            fin = True
        if fin or force:
            # cleanup
            self.part_a.cmdset.delete("cmdset_trade")
            self.part_b.cmdset.delete("cmdset_trade")
            self.part_a_offers = None
            self.part_b_offers = None
            self.part_a.scripts.stop("trade_request_timeout")
            # this will kill it also from B
            del self.part_a.ndb.tradehandler
            if self.part_b.ndb.tradehandler:
                del self.part_b.ndb.tradehandler
            return True
        return False


# trading commands (will go into CmdsetTrade, initialized by the
# CmdTrade command further down).


class CmdTradeBase(Command):
    """
    Base command for Trade commands to inherit from. Implements the
    custom parsing.
    """

    def parse(self):
        """
        Parse the relevant parts and make it easily
        available to the command
        """
        self.args = self.args.strip()
        self.tradehandler = self.caller.ndb.tradehandler
        self.part_a = self.tradehandler.part_a
        self.part_b = self.tradehandler.part_b

        self.other = self.tradehandler.get_other(self.caller)
        self.msg_other = self.tradehandler.msg_other

        self.trade_started = self.tradehandler.trade_started
        self.emote = ""
        self.str_caller = "Your trade action: %s"
        self.str_other = "%s:s trade action: " % self.caller.key + "%s"
        if ":" in self.args:
            self.args, self.emote = [part.strip() for part in self.args.rsplit(":", 1)]
            self.str_caller = 'You say, "' + self.emote + '"\n  [%s]'
            if self.caller.has_account:
                self.str_other = '|c%s|n says, "' % self.caller.key + self.emote + '"\n  [%s]'
            else:
                self.str_other = '%s says, "' % self.caller.key + self.emote + '"\n  [%s]'


# trade help


class CmdTradeHelp(CmdTradeBase):
    """
    help command for the trade system.

    Usage:
        trade help

    Displays help for the trade commands.
    """

    key = "trade help"
    locks = "cmd:all()"
    help_category = "Trade"

    def func(self):
        """Show the help"""
        string = """
        Trading commands

        |woffer <objects> [:emote]|n
            offer one or more objects for trade. The emote can be used for
            RP/arguments. A new offer will require both parties to re-accept
            it again.
        |waccept [:emote]|n
            accept the currently standing offer from both sides. Also 'agree'
            works. Once both have accepted, the deal is finished and goods
            will change hands.
        |wdecline [:emote]|n
            change your mind and remove a previous accept (until other
            has also accepted)
        |wstatus|n
            show the current offers on each side of the deal. Also 'offers'
            and 'deal' works.
        |wevaluate <nr> or <offer>|n
            examine any offer in the deal. List them with the 'status' command.
        |wend trade|n
            end the negotiations prematurely. No trade will take place.

        You can also use |wemote|n, |wsay|n etc to discuss
        without making a decision or offer.
        """
        self.caller.msg(string)


# offer


class CmdOffer(CmdTradeBase):
    """
    offer one or more items in trade.

    Usage:
      offer <object> [, object2, ...][:emote]

    Offer objects in trade. This will replace the currently
    standing offer.
    """

    key = "offer"
    locks = "cmd:all()"
    help_category = "Trading"

    def func(self):
        """implement the offer"""

        caller = self.caller
        if not self.args:
            caller.msg("Usage: offer <object> [, object2, ...] [:emote]")
            return
        if not self.trade_started:
            caller.msg("Wait until the other party has accepted to trade with you.")
            return

        # gather all offers
        offers = [part.strip() for part in self.args.split(",")]
        offerobjs = []
        for offername in offers:
            obj = caller.search(offername)
            if not obj:
                return
            offerobjs.append(obj)
        self.tradehandler.offer(self.caller, *offerobjs)

        # output
        if len(offerobjs) > 1:
            objnames = (
                ", ".join("|w%s|n" % obj.key for obj in offerobjs[:-1])
                + " and |w%s|n" % offerobjs[-1].key
            )
        else:
            objnames = "|w%s|n" % offerobjs[0].key

        caller.msg(self.str_caller % ("You offer %s" % objnames))
        self.msg_other(caller, self.str_other % ("They offer %s" % objnames))


# accept


class CmdAccept(CmdTradeBase):
    """
    accept the standing offer

    Usage:
      accept [:emote]
      agreee [:emote]

    This will accept the current offer. The other party must also accept
    for the deal to go through. You can use the 'decline' command to change
    your mind as long as the other party has not yet accepted. You can inspect
    the current offer using the 'offers' command.
    """

    key = "accept"
    aliases = ["agree"]
    locks = "cmd:all()"
    help_category = "Trading"

    def func(self):
        """accept the offer"""
        caller = self.caller
        if not self.trade_started:
            caller.msg("Wait until the other party has accepted to trade with you.")
            return
        if self.tradehandler.accept(self.caller):
            # deal finished. Trade ended and cleaned.
            caller.msg(
                self.str_caller
                % "You |gaccept|n the deal. |gDeal is made and goods changed hands.|n"
            )
            self.msg_other(
                caller,
                self.str_other % "%s |gaccepts|n the deal."
                " |gDeal is made and goods changed hands.|n" % caller.key,
            )
        else:
            # a one-sided accept.
            caller.msg(
                self.str_caller
                % "You |Gaccept|n the offer. %s must now also accept."
                % self.other.key
            )
            self.msg_other(
                caller,
                self.str_other % "%s |Gaccepts|n the offer. You must now also accept." % caller.key,
            )


# decline


class CmdDecline(CmdTradeBase):
    """
    decline the standing offer

    Usage:
      decline [:emote]

    This will decline a previously 'accept'ed offer (so this allows you to
    change your mind). You can only use this as long as the other party
    has not yet accepted the deal. Also, changing the offer will automatically
    decline the old offer.
    """

    key = "decline"
    locks = "cmd:all()"
    help_category = "Trading"

    def func(self):
        """decline the offer"""
        caller = self.caller
        if not self.trade_started:
            caller.msg("Wait until the other party has accepted to trade with you.")
            return
        offer_a, offer_b = self.tradehandler.list()
        if not offer_a or not offer_b:
            caller.msg("No offers have been made yet, so there is nothing to decline.")
            return
        if self.tradehandler.decline(self.caller):
            # changed a previous accept
            caller.msg(self.str_caller % "You change your mind, |Rdeclining|n the current offer.")
            self.msg_other(
                caller,
                self.str_other
                % "%s changes their mind, |Rdeclining|n the current offer."
                % caller.key,
            )
        else:
            # no acceptance to change
            caller.msg(self.str_caller % "You |Rdecline|n the current offer.")
            self.msg_other(caller, self.str_other % "%s declines the current offer." % caller.key)


# evaluate

# Note: This version only shows the description. If your particular game
# lists other important properties of objects (such as weapon damage, weight,
# magical properties, ammo requirements or whatnot), then you need to add this
# here.


class CmdEvaluate(CmdTradeBase):
    """
    evaluate objects on offer

    Usage:
      evaluate <offered object>

    This allows you to examine any object currently on offer, to
    determine if it's worth your while.
    """

    key = "evaluate"
    aliases = ["eval"]
    locks = "cmd:all()"
    help_category = "Trading"

    def func(self):
        """evaluate an object"""
        caller = self.caller
        if not self.args:
            caller.msg("Usage: evaluate <offered object>")
            return
        # we also accept indices
        try:
            ind = int(self.args)
            self.args = ind - 1
        except Exception:
            # not a valid index - ignore
            pass

        offer = self.tradehandler.search(self.args)
        if not offer:
            caller.msg("No offer matching '%s' was found." % self.args)
            return
        # show the description
        caller.msg(offer.db.desc)


# status


class CmdStatus(CmdTradeBase):
    """
    show a list of the current deal

    Usage:
      status
      deal
      offers

    Shows the currently suggested offers on each sides of the deal. To
    accept the current deal, use the 'accept' command. Use 'offer' to
    change your deal. You might also want to use 'say', 'emote' etc to
    try to influence the other part in the deal.
    """

    key = "status"
    aliases = ["offers", "deal"]
    locks = "cmd:all()"
    help_category = "Trading"

    def func(self):
        """Show the current deal"""
        caller = self.caller
        part_a_offers, part_b_offers = self.tradehandler.list()
        count = 1
        part_a_offerlist = []
        for offer in part_a_offers:
            part_a_offerlist.append("\n |w%i|n %s" % (count, offer.key))
            count += 1
        if not part_a_offerlist:
            part_a_offerlist = "\n <nothing>"
        part_b_offerlist = []
        for offer in part_b_offers:
            part_b_offerlist.append("\n |w%i|n %s" % (count, offer.key))
            count += 1
        if not part_b_offerlist:
            part_b_offerlist = "\n <nothing>"

        string = "|gOffered by %s:|n%s\n|yOffered by %s:|n%s" % (
            self.part_a.key,
            "".join(part_a_offerlist),
            self.part_b.key,
            "".join(part_b_offerlist),
        )
        accept_a = self.tradehandler.part_a_accepted and "|gYes|n" or "|rNo|n"
        accept_b = self.tradehandler.part_b_accepted and "|gYes|n" or "|rNo|n"
        string += "\n\n%s agreed: %s, %s agreed: %s" % (
            self.part_a.key,
            accept_a,
            self.part_b.key,
            accept_b,
        )
        string += "\n Use 'offer', 'eval' and 'accept'/'decline' to trade. See also 'trade help'."
        caller.msg(string)


# finish


class CmdFinish(CmdTradeBase):
    """
    end the trade prematurely

    Usage:
      end trade [:say]
      finish trade [:say]

    This ends the trade prematurely. No trade will take place.

    """

    key = "end trade"
    aliases = "finish trade"
    locks = "cmd:all()"
    help_category = "Trading"

    def func(self):
        """end trade"""
        caller = self.caller
        self.tradehandler.finish(force=True)
        caller.msg(self.str_caller % "You |raborted|n trade. No deal was made.")
        self.msg_other(
            caller, self.str_other % "%s |raborted|n trade. No deal was made." % caller.key
        )


# custom Trading cmdset


class CmdsetTrade(CmdSet):
    """
    This cmdset is added when trade is initated. It is handled by the
    trade event handler.
    """

    key = "cmdset_trade"

    def at_cmdset_creation(self):
        """Called when cmdset is created"""
        self.add(CmdTradeHelp())
        self.add(CmdOffer())
        self.add(CmdAccept())
        self.add(CmdDecline())
        self.add(CmdEvaluate())
        self.add(CmdStatus())
        self.add(CmdFinish())


# access command - once both have given this, this will create the
# trading cmdset to start trade.


class CmdTrade(Command):
    """
    Initiate trade with another party

    Usage:
      trade <other party> [:say]
      trade <other party> accept [:say]
      trade <other party> decline [:say]

    Initiate trade with another party. The other party needs to repeat
    this command with trade accept/decline within a minute in order to
    properly initiate the trade action. You can use the decline option
    yourself if you want to retract an already suggested trade. The
    optional say part works like the say command and allows you to add
    info to your choice.
    """

    key = "trade"
    aliases = ["barter"]
    locks = "cmd:all()"
    help_category = "General"

    def func(self):
        """Initiate trade"""

        if not self.args:
            if self.caller.ndb.tradehandler and self.caller.ndb.tradeevent.trade_started:
                self.caller.msg("You are already in a trade. Use 'end trade' to abort it.")
            else:
                self.caller.msg("Usage: trade <other party> [accept|decline] [:emote]")
            return
        self.args = self.args.strip()

        # handle the emote manually here
        selfemote = ""
        theiremote = ""
        if ":" in self.args:
            self.args, emote = [part.strip() for part in self.args.rsplit(":", 1)]
            selfemote = 'You say, "%s"\n  ' % emote
            if self.caller.has_account:
                theiremote = '|c%s|n says, "%s"\n  ' % (self.caller.key, emote)
            else:
                theiremote = '%s says, "%s"\n  ' % (self.caller.key, emote)

        # for the sake of this command, the caller is always part_a; this
        # might not match the actual name in tradehandler (in the case of
        # using this command to accept/decline a trade invitation).
        part_a = self.caller
        accept = "accept" in self.args
        decline = "decline" in self.args
        if accept:
            part_b = self.args.rstrip("accept").strip()
        elif decline:
            part_b = self.args.rstrip("decline").strip()
        else:
            part_b = self.args
        part_b = self.caller.search(part_b)
        if not part_b:
            return
        if part_a == part_b:
            part_a.msg("You play trader with yourself.")
            return

        # messages
        str_init_a = "You ask to trade with %s. They need to accept within %s secs."
        str_init_b = "%s wants to trade with you. Use |wtrade %s accept/decline [:emote]|n to answer (within %s secs)."
        str_noinit_a = "%s declines the trade"
        str_noinit_b = "You decline trade with %s."
        str_start_a = "%s starts to trade with you. See |wtrade help|n for aid."
        str_start_b = "You start to trade with %s. See |wtrade help|n for aid."

        if not (accept or decline):
            # initialization of trade
            if self.caller.ndb.tradehandler:
                # trying to start trade without stopping a previous one
                if self.caller.ndb.tradehandler.trade_started:
                    string = "You are already in trade with %s. You need to end trade first."
                else:
                    string = "You are already trying to initiate trade with %s. You need to decline that trade first."
                self.caller.msg(string % part_b.key)
            elif part_b.ndb.tradehandler and part_b.ndb.tradehandler.part_b == part_a:
                # this is equivalent to part_a accepting a trade from part_b (so roles are reversed)
                part_b.ndb.tradehandler.join(part_a)
                part_b.msg(theiremote + str_start_a % part_a.key)
                part_a.msg(selfemote + str_start_b % part_b.key)
            else:
                # initiate a new trade
                TradeHandler(part_a, part_b)
                part_a.msg(selfemote + str_init_a % (part_b.key, TRADE_TIMEOUT))
                part_b.msg(theiremote + str_init_b % (part_a.key, part_a.key, TRADE_TIMEOUT))
                part_a.scripts.add(TradeTimeout)
            return
        elif accept:
            # accept a trade proposal from part_b (so roles are reversed)
            if part_a.ndb.tradehandler:
                # already in a trade
                part_a.msg(
                    "You are already in trade with %s. You need to end that first." % part_b.key
                )
                return
            if part_b.ndb.tradehandler.join(part_a):
                part_b.msg(theiremote + str_start_a % part_a.key)
                part_a.msg(selfemote + str_start_b % part_b.key)
            else:
                part_a.msg("No trade proposal to accept.")
            return
        else:
            # decline trade proposal from part_b (so roles are reversed)
            if part_a.ndb.tradehandler and part_a.ndb.tradehandler.part_b == part_a:
                # stopping an invite
                part_a.ndb.tradehandler.finish(force=True)
                part_b.msg(theiremote + "%s aborted trade attempt with you." % part_a)
                part_a.msg(selfemote + "You aborted the trade attempt with %s." % part_b)
            elif part_b.ndb.tradehandler and part_b.ndb.tradehandler.unjoin(part_a):
                part_b.msg(theiremote + str_noinit_a % part_a.key)
                part_a.msg(selfemote + str_noinit_b % part_b.key)
            else:
                part_a.msg("No trade proposal to decline.")
            return



"""
Godot Websocket - ChrisLR 2022

This file contains the necessary code and data to convert text with color tags to bbcode (For godot)
"""
from evennia.utils.ansi import *
from evennia.utils.text2html import TextToHTMLparser

# All xterm256 RGB equivalents

XTERM256_FG = "\033[38;5;{}m"
XTERM256_BG = "\033[48;5;{}m"

COLOR_INDICE_TO_HEX = {
    "color-000": "#000000",
    "color-001": "#800000",
    "color-002": "#008000",
    "color-003": "#808000",
    "color-004": "#000080",
    "color-005": "#800080",
    "color-006": "#008080",
    "color-007": "#c0c0c0",
    "color-008": "#808080",
    "color-009": "#ff0000",
    "color-010": "#00ff00",
    "color-011": "#ffff00",
    "color-012": "#0000ff",
    "color-013": "#ff00ff",
    "color-014": "#00ffff",
    "color-015": "#ffffff",
    "color-016": "#000000",
    "color-017": "#00005f",
    "color-018": "#000087",
    "color-019": "#0000af",
    "color-020": "#0000df",
    "color-021": "#0000ff",
    "color-022": "#005f00",
    "color-023": "#005f5f",
    "color-024": "#005f87",
    "color-025": "#005faf",
    "color-026": "#005fdf",
    "color-027": "#005fff",
    "color-028": "#008700",
    "color-029": "#00875f",
    "color-030": "#008787",
    "color-031": "#0087af",
    "color-032": "#0087df",
    "color-033": "#0087ff",
    "color-034": "#00af00",
    "color-035": "#00af5f",
    "color-036": "#00af87",
    "color-037": "#00afaf",
    "color-038": "#00afdf",
    "color-039": "#00afff",
    "color-040": "#00df00",
    "color-041": "#00df5f",
    "color-042": "#00df87",
    "color-043": "#00dfaf",
    "color-044": "#00dfdf",
    "color-045": "#00dfff",
    "color-046": "#00ff00",
    "color-047": "#00ff5f",
    "color-048": "#00ff87",
    "color-049": "#00ffaf",
    "color-050": "#00ffdf",
    "color-051": "#00ffff",
    "color-052": "#5f0000",
    "color-053": "#5f005f",
    "color-054": "#5f0087",
    "color-055": "#5f00af",
    "color-056": "#5f00df",
    "color-057": "#5f00ff",
    "color-058": "#5f5f00",
    "color-059": "#5f5f5f",
    "color-060": "#5f5f87",
    "color-061": "#5f5faf",
    "color-062": "#5f5fdf",
    "color-063": "#5f5fff",
    "color-064": "#5f8700",
    "color-065": "#5f875f",
    "color-066": "#5f8787",
    "color-067": "#5f87af",
    "color-068": "#5f87df",
    "color-069": "#5f87ff",
    "color-070": "#5faf00",
    "color-071": "#5faf5f",
    "color-072": "#5faf87",
    "color-073": "#5fafaf",
    "color-074": "#5fafdf",
    "color-075": "#5fafff",
    "color-076": "#5fdf00",
    "color-077": "#5fdf5f",
    "color-078": "#5fdf87",
    "color-079": "#5fdfaf",
    "color-080": "#5fdfdf",
    "color-081": "#5fdfff",
    "color-082": "#5fff00",
    "color-083": "#5fff5f",
    "color-084": "#5fff87",
    "color-085": "#5fffaf",
    "color-086": "#5fffdf",
    "color-087": "#5fffff",
    "color-088": "#870000",
    "color-089": "#87005f",
    "color-090": "#870087",
    "color-091": "#8700af",
    "color-092": "#8700df",
    "color-093": "#8700ff",
    "color-094": "#875f00",
    "color-095": "#875f5f",
    "color-096": "#875f87",
    "color-097": "#875faf",
    "color-098": "#875fdf",
    "color-099": "#875fff",
    "color-100": "#878700",
    "color-101": "#87875f",
    "color-102": "#878787",
    "color-103": "#8787af",
    "color-104": "#8787df",
    "color-105": "#8787ff",
    "color-106": "#87af00",
    "color-107": "#87af5f",
    "color-108": "#87af87",
    "color-109": "#87afaf",
    "color-110": "#87afdf",
    "color-111": "#87afff",
    "color-112": "#87df00",
    "color-113": "#87df5f",
    "color-114": "#87df87",
    "color-115": "#87dfaf",
    "color-116": "#87dfdf",
    "color-117": "#87dfff",
    "color-118": "#87ff00",
    "color-119": "#87ff5f",
    "color-120": "#87ff87",
    "color-121": "#87ffaf",
    "color-122": "#87ffdf",
    "color-123": "#87ffff",
    "color-124": "#af0000",
    "color-125": "#af005f",
    "color-126": "#af0087",
    "color-127": "#af00af",
    "color-128": "#af00df",
    "color-129": "#af00ff",
    "color-130": "#af5f00",
    "color-131": "#af5f5f",
    "color-132": "#af5f87",
    "color-133": "#af5faf",
    "color-134": "#af5fdf",
    "color-135": "#af5fff",
    "color-136": "#af8700",
    "color-137": "#af875f",
    "color-138": "#af8787",
    "color-139": "#af87af",
    "color-140": "#af87df",
    "color-141": "#af87ff",
    "color-142": "#afaf00",
    "color-143": "#afaf5f",
    "color-144": "#afaf87",
    "color-145": "#afafaf",
    "color-146": "#afafdf",
    "color-147": "#afafff",
    "color-148": "#afdf00",
    "color-149": "#afdf5f",
    "color-150": "#afdf87",
    "color-151": "#afdfaf",
    "color-152": "#afdfdf",
    "color-153": "#afdfff",
    "color-154": "#afff00",
    "color-155": "#afff5f",
    "color-156": "#afff87",
    "color-157": "#afffaf",
    "color-158": "#afffdf",
    "color-159": "#afffff",
    "color-160": "#df0000",
    "color-161": "#df005f",
    "color-162": "#df0087",
    "color-163": "#df00af",
    "color-164": "#df00df",
    "color-165": "#df00ff",
    "color-166": "#df5f00",
    "color-167": "#df5f5f",
    "color-168": "#df5f87",
    "color-169": "#df5faf",
    "color-170": "#df5fdf",
    "color-171": "#df5fff",
    "color-172": "#df8700",
    "color-173": "#df875f",
    "color-174": "#df8787",
    "color-175": "#df87af",
    "color-176": "#df87df",
    "color-177": "#df87ff",
    "color-178": "#dfaf00",
    "color-179": "#dfaf5f",
    "color-180": "#dfaf87",
    "color-181": "#dfafaf",
    "color-182": "#dfafdf",
    "color-183": "#dfafff",
    "color-184": "#dfdf00",
    "color-185": "#dfdf5f",
    "color-186": "#dfdf87",
    "color-187": "#dfdfaf",
    "color-188": "#dfdfdf",
    "color-189": "#dfdfff",
    "color-190": "#dfff00",
    "color-191": "#dfff5f",
    "color-192": "#dfff87",
    "color-193": "#dfffaf",
    "color-194": "#dfffdf",
    "color-195": "#dfffff",
    "color-196": "#ff0000",
    "color-197": "#ff005f",
    "color-198": "#ff0087",
    "color-199": "#ff00af",
    "color-200": "#ff00df",
    "color-201": "#ff00ff",
    "color-202": "#ff5f00",
    "color-203": "#ff5f5f",
    "color-204": "#ff5f87",
    "color-205": "#ff5faf",
    "color-206": "#ff5fdf",
    "color-207": "#ff5fff",
    "color-208": "#ff8700",
    "color-209": "#ff875f",
    "color-210": "#ff8787",
    "color-211": "#ff87af",
    "color-212": "#ff87df",
    "color-213": "#ff87ff",
    "color-214": "#ffaf00",
    "color-215": "#ffaf5f",
    "color-216": "#ffaf87",
    "color-217": "#ffafaf",
    "color-218": "#ffafdf",
    "color-219": "#ffafff",
    "color-220": "#ffdf00",
    "color-221": "#ffdf5f",
    "color-222": "#ffdf87",
    "color-223": "#ffdfaf",
    "color-224": "#ffdfdf",
    "color-225": "#ffdfff",
    "color-226": "#ffff00",
    "color-227": "#ffff5f",
    "color-228": "#ffff87",
    "color-229": "#ffffaf",
    "color-230": "#ffffdf",
    "color-231": "#ffffff",
    "color-232": "#080808",
    "color-233": "#121212",
    "color-234": "#1c1c1c",
    "color-235": "#262626",
    "color-236": "#303030",
    "color-237": "#3a3a3a",
    "color-238": "#444444",
    "color-239": "#4e4e4e",
    "color-240": "#585858",
    "color-241": "#606060",
    "color-242": "#666666",
    "color-243": "#767676",
    "color-244": "#808080",
    "color-245": "#8a8a8a",
    "color-246": "#949494",
    "color-247": "#9e9e9e",
    "color-248": "#a8a8a8",
    "color-249": "#b2b2b2",
    "color-250": "#bcbcbc",
    "color-251": "#c6c6c6",
    "color-252": "#d0d0d0",
    "color-253": "#dadada",
    "color-254": "#e4e4e4",
    "color-255": "#eeeeee",
    "bgcolor-000": "#000000",
    "bgcolor-001": "#800000",
    "bgcolor-002": "#008000",
    "bgcolor-003": "#808000",
    "bgcolor-004": "#000080",
    "bgcolor-005": "#800080",
    "bgcolor-006": "#008080",
    "bgcolor-007": "#c0c0c0",
    "bgcolor-008": "#808080",
    "bgcolor-009": "#ff0000",
    "bgcolor-010": "#00ff00",
    "bgcolor-011": "#ffff00",
    "bgcolor-012": "#0000ff",
    "bgcolor-013": "#ff00ff",
    "bgcolor-014": "#00ffff",
    "bgcolor-015": "#ffffff",
    "bgcolor-016": "#000000",
    "bgcolor-017": "#00005f",
    "bgcolor-018": "#000087",
    "bgcolor-019": "#0000af",
    "bgcolor-020": "#0000df",
    "bgcolor-021": "#0000ff",
    "bgcolor-022": "#005f00",
    "bgcolor-023": "#005f5f",
    "bgcolor-024": "#005f87",
    "bgcolor-025": "#005faf",
    "bgcolor-026": "#005fdf",
    "bgcolor-027": "#005fff",
    "bgcolor-028": "#008700",
    "bgcolor-029": "#00875f",
    "bgcolor-030": "#008787",
    "bgcolor-031": "#0087af",
    "bgcolor-032": "#0087df",
    "bgcolor-033": "#0087ff",
    "bgcolor-034": "#00af00",
    "bgcolor-035": "#00af5f",
    "bgcolor-036": "#00af87",
    "bgcolor-037": "#00afaf",
    "bgcolor-038": "#00afdf",
    "bgcolor-039": "#00afff",
    "bgcolor-040": "#00df00",
    "bgcolor-041": "#00df5f",
    "bgcolor-042": "#00df87",
    "bgcolor-043": "#00dfaf",
    "bgcolor-044": "#00dfdf",
    "bgcolor-045": "#00dfff",
    "bgcolor-046": "#00ff00",
    "bgcolor-047": "#00ff5f",
    "bgcolor-048": "#00ff87",
    "bgcolor-049": "#00ffaf",
    "bgcolor-050": "#00ffdf",
    "bgcolor-051": "#00ffff",
    "bgcolor-052": "#5f0000",
    "bgcolor-053": "#5f005f",
    "bgcolor-054": "#5f0087",
    "bgcolor-055": "#5f00af",
    "bgcolor-056": "#5f00df",
    "bgcolor-057": "#5f00ff",
    "bgcolor-058": "#5f5f00",
    "bgcolor-059": "#5f5f5f",
    "bgcolor-060": "#5f5f87",
    "bgcolor-061": "#5f5faf",
    "bgcolor-062": "#5f5fdf",
    "bgcolor-063": "#5f5fff",
    "bgcolor-064": "#5f8700",
    "bgcolor-065": "#5f875f",
    "bgcolor-066": "#5f8787",
    "bgcolor-067": "#5f87af",
    "bgcolor-068": "#5f87df",
    "bgcolor-069": "#5f87ff",
    "bgcolor-070": "#5faf00",
    "bgcolor-071": "#5faf5f",
    "bgcolor-072": "#5faf87",
    "bgcolor-073": "#5fafaf",
    "bgcolor-074": "#5fafdf",
    "bgcolor-075": "#5fafff",
    "bgcolor-076": "#5fdf00",
    "bgcolor-077": "#5fdf5f",
    "bgcolor-078": "#5fdf87",
    "bgcolor-079": "#5fdfaf",
    "bgcolor-080": "#5fdfdf",
    "bgcolor-081": "#5fdfff",
    "bgcolor-082": "#5fff00",
    "bgcolor-083": "#5fff5f",
    "bgcolor-084": "#5fff87",
    "bgcolor-085": "#5fffaf",
    "bgcolor-086": "#5fffdf",
    "bgcolor-087": "#5fffff",
    "bgcolor-088": "#870000",
    "bgcolor-089": "#87005f",
    "bgcolor-090": "#870087",
    "bgcolor-091": "#8700af",
    "bgcolor-092": "#8700df",
    "bgcolor-093": "#8700ff",
    "bgcolor-094": "#875f00",
    "bgcolor-095": "#875f5f",
    "bgcolor-096": "#875f87",
    "bgcolor-097": "#875faf",
    "bgcolor-098": "#875fdf",
    "bgcolor-099": "#875fff",
    "bgcolor-100": "#878700",
    "bgcolor-101": "#87875f",
    "bgcolor-102": "#878787",
    "bgcolor-103": "#8787af",
    "bgcolor-104": "#8787df",
    "bgcolor-105": "#8787ff",
    "bgcolor-106": "#87af00",
    "bgcolor-107": "#87af5f",
    "bgcolor-108": "#87af87",
    "bgcolor-109": "#87afaf",
    "bgcolor-110": "#87afdf",
    "bgcolor-111": "#87afff",
    "bgcolor-112": "#87df00",
    "bgcolor-113": "#87df5f",
    "bgcolor-114": "#87df87",
    "bgcolor-115": "#87dfaf",
    "bgcolor-116": "#87dfdf",
    "bgcolor-117": "#87dfff",
    "bgcolor-118": "#87ff00",
    "bgcolor-119": "#87ff5f",
    "bgcolor-120": "#87ff87",
    "bgcolor-121": "#87ffaf",
    "bgcolor-122": "#87ffdf",
    "bgcolor-123": "#87ffff",
    "bgcolor-124": "#af0000",
    "bgcolor-125": "#af005f",
    "bgcolor-126": "#af0087",
    "bgcolor-127": "#af00af",
    "bgcolor-128": "#af00df",
    "bgcolor-129": "#af00ff",
    "bgcolor-130": "#af5f00",
    "bgcolor-131": "#af5f5f",
    "bgcolor-132": "#af5f87",
    "bgcolor-133": "#af5faf",
    "bgcolor-134": "#af5fdf",
    "bgcolor-135": "#af5fff",
    "bgcolor-136": "#af8700",
    "bgcolor-137": "#af875f",
    "bgcolor-138": "#af8787",
    "bgcolor-139": "#af87af",
    "bgcolor-140": "#af87df",
    "bgcolor-141": "#af87ff",
    "bgcolor-142": "#afaf00",
    "bgcolor-143": "#afaf5f",
    "bgcolor-144": "#afaf87",
    "bgcolor-145": "#afafaf",
    "bgcolor-146": "#afafdf",
    "bgcolor-147": "#afafff",
    "bgcolor-148": "#afdf00",
    "bgcolor-149": "#afdf5f",
    "bgcolor-150": "#afdf87",
    "bgcolor-151": "#afdfaf",
    "bgcolor-152": "#afdfdf",
    "bgcolor-153": "#afdfff",
    "bgcolor-154": "#afff00",
    "bgcolor-155": "#afff5f",
    "bgcolor-156": "#afff87",
    "bgcolor-157": "#afffaf",
    "bgcolor-158": "#afffdf",
    "bgcolor-159": "#afffff",
    "bgcolor-160": "#df0000",
    "bgcolor-161": "#df005f",
    "bgcolor-162": "#df0087",
    "bgcolor-163": "#df00af",
    "bgcolor-164": "#df00df",
    "bgcolor-165": "#df00ff",
    "bgcolor-166": "#df5f00",
    "bgcolor-167": "#df5f5f",
    "bgcolor-168": "#df5f87",
    "bgcolor-169": "#df5faf",
    "bgcolor-170": "#df5fdf",
    "bgcolor-171": "#df5fff",
    "bgcolor-172": "#df8700",
    "bgcolor-173": "#df875f",
    "bgcolor-174": "#df8787",
    "bgcolor-175": "#df87af",
    "bgcolor-176": "#df87df",
    "bgcolor-177": "#df87ff",
    "bgcolor-178": "#dfaf00",
    "bgcolor-179": "#dfaf5f",
    "bgcolor-180": "#dfaf87",
    "bgcolor-181": "#dfafaf",
    "bgcolor-182": "#dfafdf",
    "bgcolor-183": "#dfafff",
    "bgcolor-184": "#dfdf00",
    "bgcolor-185": "#dfdf5f",
    "bgcolor-186": "#dfdf87",
    "bgcolor-187": "#dfdfaf",
    "bgcolor-188": "#dfdfdf",
    "bgcolor-189": "#dfdfff",
    "bgcolor-190": "#dfff00",
    "bgcolor-191": "#dfff5f",
    "bgcolor-192": "#dfff87",
    "bgcolor-193": "#dfffaf",
    "bgcolor-194": "#dfffdf",
    "bgcolor-195": "#dfffff",
    "bgcolor-196": "#ff0000",
    "bgcolor-197": "#ff005f",
    "bgcolor-198": "#ff0087",
    "bgcolor-199": "#ff00af",
    "bgcolor-200": "#ff00df",
    "bgcolor-201": "#ff00ff",
    "bgcolor-202": "#ff5f00",
    "bgcolor-203": "#ff5f5f",
    "bgcolor-204": "#ff5f87",
    "bgcolor-205": "#ff5faf",
    "bgcolor-206": "#ff5fdf",
    "bgcolor-207": "#ff5fff",
    "bgcolor-208": "#ff8700",
    "bgcolor-209": "#ff875f",
    "bgcolor-210": "#ff8787",
    "bgcolor-211": "#ff87af",
    "bgcolor-212": "#ff87df",
    "bgcolor-213": "#ff87ff",
    "bgcolor-214": "#ffaf00",
    "bgcolor-215": "#ffaf5f",
    "bgcolor-216": "#ffaf87",
    "bgcolor-217": "#ffafaf",
    "bgcolor-218": "#ffafdf",
    "bgcolor-219": "#ffafff",
    "bgcolor-220": "#ffdf00",
    "bgcolor-221": "#ffdf5f",
    "bgcolor-222": "#ffdf87",
    "bgcolor-223": "#ffdfaf",
    "bgcolor-224": "#ffdfdf",
    "bgcolor-225": "#ffdfff",
    "bgcolor-226": "#ffff00",
    "bgcolor-227": "#ffff5f",
    "bgcolor-228": "#ffff87",
    "bgcolor-229": "#ffffaf",
    "bgcolor-230": "#ffffdf",
    "bgcolor-231": "#ffffff",
    "bgcolor-232": "#080808",
    "bgcolor-233": "#121212",
    "bgcolor-234": "#1c1c1c",
    "bgcolor-235": "#262626",
    "bgcolor-236": "#303030",
    "bgcolor-237": "#3a3a3a",
    "bgcolor-238": "#444444",
    "bgcolor-239": "#4e4e4e",
    "bgcolor-240": "#585858",
    "bgcolor-241": "#606060",
    "bgcolor-242": "#666666",
    "bgcolor-243": "#767676",
    "bgcolor-244": "#808080",
    "bgcolor-245": "#8a8a8a",
    "bgcolor-246": "#949494",
    "bgcolor-247": "#9e9e9e",
    "bgcolor-248": "#a8a8a8",
    "bgcolor-249": "#b2b2b2",
    "bgcolor-250": "#bcbcbc",
    "bgcolor-251": "#c6c6c6",
    "bgcolor-252": "#d0d0d0",
    "bgcolor-253": "#dadada",
    "bgcolor-254": "#e4e4e4",
    "bgcolor-255": "#eeeeee",
}


"""
The classes below exist to properly encapsulate text and other tag classes
because the order of how tags are opened and closed are important to display in godot.
"""


class RootTag:
    """
    The Root tag class made to contain other tags.
    """

    __slots__ = ("child",)

    def __init__(self):
        self.child = None

    def __str__(self):
        return str(self.child) if self.child else ""


class ChildTag:
    """
    A node made to be contained.
    """

    def __init__(self, parent):
        self.parent = parent
        if parent:
            parent.child = self

    def set_parent(self, parent):
        self.parent = parent
        if parent:
            parent.child = self


class TextTag(ChildTag):
    """
    A BBCodeTag node to output regular text.
    Output: SomeText
    """

    __slots__ = ("parent", "child", "text")

    def __init__(self, parent, text):
        super().__init__(parent)
        self.text = text
        self.child = None

    def __str__(self):
        return f"{self.text}{self.child or ''}"


class BBCodeTag(ChildTag):
    """
    Base BBCodeTag node to encapsulate and be encapsulated.
    """

    __slots__ = (
        "parent",
        "child",
    )

    code = ""

    def __init__(self, parent):
        super().__init__(parent)
        self.child = None

    def __str__(self):
        return f"[{self.code}]{self.child or ''}[/{self.code}]"


class UnderlineTag(BBCodeTag):
    """
    A BBCodeTag node for underlined text.
    Output: [u]Underlined Text[/u]
    """

    code = "u"


class BlinkTag(BBCodeTag):
    """
    A BBCodeTag node for blinking text.
    Output: [blink]Blinking Text[/blink]
    """

    code = "blink"


class ColorTag(BBCodeTag):
    """
    A BBCodeTag node for foreground color.
    Output: [fgcolor=#000000]Colorized Text[/fgcolor]
    """

    __slots__ = (
        "parent",
        "child",
        "color_hex",
    )

    code = "color"

    def __init__(self, parent, color_hex):
        super().__init__(parent)
        self.color_hex = color_hex

    def __str__(self):
        return f"[{self.code}={self.color_hex}]{self.child or ''}[/{self.code}]"


class BGColorTag(ColorTag):
    """
    A BBCodeTag node for background color.
    Output: [bgcolor=#000000]Colorized Text[/bgcolor]
    """

    code = "bgcolor"


class UrlTag(BBCodeTag):
    """
    A BBCodeTag node used for urls.
    Output: [url=www.example.com]Child Text[/url]

    """

    __slots__ = (
        "parent",
        "child",
        "url_data",
    )

    code = "url"

    def __init__(self, parent, url_data=""):
        super().__init__(parent)
        self.url_data = url_data

    def __str__(self):
        return f"[{self.code}={self.url_data}]{self.child or ''}[/{self.code}]"


class TextToBBCODEparser(TextToHTMLparser):
    """
    This class describes a parser for converting from ANSI to BBCode.
    It inherits from the TextToHTMLParser and overrides the specifics for bbcode.
    """

    def convert_urls(self, text):
        """
        Converts urls within text to bbcode style

        Args:
            text (str): Text to parse

        Returns:
             text (str): Processed text
        """
        # Converts to bbcode styled urls
        return self.re_url.sub(r"[url=\1]\1[/url]\2", text)

    def sub_mxp_links(self, match):
        """
        Helper method to be passed to re.sub,
        replaces MXP links with bbcode.

        Args:
            match (re.Matchobject): Match for substitution.

        Returns:
            text (str): Processed text.

        """
        cmd, text = [grp.replace('"', "\\&quot;") for grp in match.groups()]
        val = f"[mxp=send cmd={cmd}]{text}[/mxp]"

        return val

    def sub_mxp_urls(self, match):
        """
        Helper method to be passed to re.sub,
        replaces MXP links with bbcode.
        Args:
            match (re.Matchobject): Match for substitution.
        Returns:
            text (str): Processed text.
        """

        url, text = [grp.replace('"', "\\&quot;") for grp in match.groups()]
        val = f"[url={url}]{text}[/url]"

        return val

    def sub_text(self, match):
        """
        Helper method to be passed to re.sub,
        for handling all substitutions.

        Args:
            match (re.Matchobject): Match for substitution.

        Returns:
            text (str): Processed text.

        """
        cdict = match.groupdict()
        if cdict["lineend"]:
            return "\n"

        return None

    def format_styles(self, text):
        """
        Takes a string with parsed ANSI codes and replaces them with bbcode style tags

        Args:
            text (str): The string to process.

        Returns:
            text (str): Processed text.
        """

        # split out the ANSI codes and clean out any empty items
        str_list = [substr for substr in self.re_style.split(text) if substr]

        inverse = False
        # default color is light grey - unhilite + white
        hilight = ANSI_UNHILITE
        fg = ANSI_WHITE
        # default bg is black
        bg = ANSI_BACK_BLACK
        previous_fg = None
        previous_bg = None
        blink = False
        underline = False
        new_style = False

        parts = []
        root_tag = RootTag()
        current_tag = root_tag

        for i, substr in enumerate(str_list):
            # reset all current styling
            if substr == ANSI_NORMAL:
                # close any existing span if necessary
                parts.append(str(root_tag))
                root_tag = RootTag()
                current_tag = root_tag
                # reset to defaults
                inverse = False
                hilight = ANSI_UNHILITE
                fg = ANSI_WHITE
                bg = ANSI_BACK_BLACK
                previous_fg = None
                previous_bg = None
                blink = False
                underline = False
                new_style = False

            # change color
            elif substr in self.ansi_color_codes + self.xterm_fg_codes:
                # set new color
                fg = substr
                new_style = True

            # change bg color
            elif substr in self.ansi_bg_codes + self.xterm_bg_codes:
                # set new bg
                bg = substr
                new_style = True

            # non-color codes
            elif substr in self.style_codes:
                new_style = True

                # hilight codes
                if substr in (ANSI_HILITE, ANSI_UNHILITE, ANSI_INV_HILITE, ANSI_INV_BLINK_HILITE):
                    # set new hilight status
                    hilight = ANSI_UNHILITE if substr == ANSI_UNHILITE else ANSI_HILITE

                # inversion codes
                if substr in (ANSI_INVERSE, ANSI_INV_HILITE, ANSI_INV_BLINK_HILITE):
                    inverse = True

                # blink codes
                if substr in (ANSI_BLINK, ANSI_BLINK_HILITE, ANSI_INV_BLINK_HILITE) and not blink:
                    blink = True
                    current_tag = BlinkTag(current_tag)

                # underline
                if substr == ANSI_UNDERLINE and not underline:
                    underline = True
                    current_tag = UnderlineTag(current_tag)

            else:
                close_tags = False
                color_tag = None
                bgcolor_tag = None
                # normal text, add text back to list
                if new_style:
                    # prior entry was cleared, which means style change
                    # get indices for the fg and bg codes
                    bg_index = self.bglist.index(bg)
                    try:
                        color_index = self.colorlist.index(hilight + fg)
                    except ValueError:
                        # xterm256 colors don't have the hilight codes
                        color_index = self.colorlist.index(fg)

                    if inverse:
                        # inverse means swap fg and bg indices
                        bg_class = "bgcolor-{}".format(str(color_index).rjust(3, "0"))
                        color_class = "color-{}".format(str(bg_index).rjust(3, "0"))
                    else:
                        # use fg and bg indices for classes
                        bg_class = "bgcolor-{}".format(str(bg_index).rjust(3, "0"))
                        color_class = "color-{}".format(str(color_index).rjust(3, "0"))

                    # black bg is the default, don't explicitly style
                    if bg_class != "bgcolor-000":
                        color_hex = COLOR_INDICE_TO_HEX.get(bg_class)
                        bgcolor_tag = BGColorTag(None, color_hex=color_hex)
                        if previous_bg and previous_bg != color_hex:
                            close_tags = True
                        else:
                            previous_bg = color_hex

                    # light grey text is the default, don't explicitly style
                    if color_class != "color-007":
                        color_hex = COLOR_INDICE_TO_HEX.get(color_class)
                        color_tag = ColorTag(None, color_hex=color_hex)
                        if previous_fg and previous_fg != color_hex:
                            close_tags = True
                        else:
                            previous_fg = color_hex

                new_tag = TextTag(None, substr)
                if close_tags:
                    # Because the order is important, we need to close the tags and reopen those who shouldn't reset.
                    new_style = False
                    parts.append(str(root_tag))
                    root_tag = RootTag()
                    current_tag = root_tag
                    if blink:
                        current_tag = BlinkTag(current_tag)

                    if underline:
                        current_tag = UnderlineTag(current_tag)

                    if bgcolor_tag:
                        bgcolor_tag.set_parent(current_tag)
                        current_tag = bgcolor_tag

                    if color_tag:
                        color_tag.set_parent(current_tag)
                        current_tag = color_tag

                    new_tag.set_parent(current_tag)
                    current_tag = new_tag
                else:
                    if bgcolor_tag:
                        bgcolor_tag.set_parent(current_tag)
                        current_tag = bgcolor_tag

                    if color_tag:
                        color_tag.set_parent(current_tag)
                        current_tag = color_tag

                    new_tag.set_parent(current_tag)
                    current_tag = new_tag

        any_text = self._get_text_tag(root_tag)
        if any_text:
            # Only append tags if text was added.
            last_part = str(root_tag)
            parts.append(last_part)

        # recombine back into string
        return "".join(parts)

    def _get_text_tag(self, root):
        child = root.child
        while child:
            if isinstance(child, TextTag):
                return child
            else:
                child = child.child

        return None

    def parse(self, text, strip_ansi=False):
        """
        Main access function, converts a text containing ANSI codes
        into html statements.

        Args:
            text (str): Text to process.
            strip_ansi (bool, optional):

        Returns:
            text (str): Parsed text.

        """
        # parse everything to ansi first
        text = parse_ansi(text, strip_ansi=strip_ansi, xterm256=True, mxp=True)
        # convert all ansi to html
        result = re.sub(self.re_string, self.sub_text, text)
        result = re.sub(self.re_mxplink, self.sub_mxp_links, result)
        result = re.sub(self.re_mxpurl, self.sub_mxp_urls, result)
        result = self.remove_bells(result)
        result = self.format_styles(result)
        result = self.remove_backspaces(result)
        result = self.convert_urls(result)

        return result


BBCODE_PARSER = TextToBBCODEparser()


#
# Access function
#


def parse_to_bbcode(string, strip_ansi=False, parser=BBCODE_PARSER):
    """
    Parses a string, replace ANSI markup with bbcode
    """
    return parser.parse(string, strip_ansi=strip_ansi)



"""
Communication commands:

- channel
- page
- irc/rss/grapevine/discord linking

"""

from django.conf import settings

from evennia.accounts import bots
from evennia.accounts.models import AccountDB
from evennia.comms.comms import DefaultChannel
from evennia.comms.models import Msg
from evennia.locks.lockhandler import LockException
from evennia.utils import create, logger, search, utils
from evennia.utils.evmenu import ask_yes_no
from evennia.utils.logger import tail_log_file
from evennia.utils.utils import class_from_module, strip_unsafe_input

COMMAND_DEFAULT_CLASS = class_from_module(settings.COMMAND_DEFAULT_CLASS)
CHANNEL_DEFAULT_TYPECLASS = class_from_module(
    settings.BASE_CHANNEL_TYPECLASS, fallback=settings.FALLBACK_CHANNEL_TYPECLASS
)


# limit symbol import for API
__all__ = (
    "CmdChannel",
    "CmdObjectChannel",
    "CmdPage",
    "CmdIRC2Chan",
    "CmdIRCStatus",
    "CmdRSS2Chan",
    "CmdGrapevine2Chan",
    "CmdDiscord2Chan",
)
_DEFAULT_WIDTH = settings.CLIENT_DEFAULT_WIDTH

# helper functions to make it easier to override the main CmdChannel
# command and to keep the legacy addcom etc commands around.


class CmdChannel(COMMAND_DEFAULT_CLASS):
    """
    Use and manage in-game channels.

    Usage:
      channel channelname <msg>
      channel channel name = <msg>
      channel        (show all subscription)
      channel/all    (show available channels)
      channel/alias channelname = alias[;alias...]
      channel/unalias alias
      channel/who channelname
      channel/history channelname [= index]
      channel/sub channelname [= alias[;alias...]]
      channel/unsub channelname[,channelname, ...]
      channel/mute channelname[,channelname,...]
      channel/unmute channelname[,channelname,...]

      channel/create channelname[;alias;alias[:typeclass]] [= description]
      channel/destroy channelname [= reason]
      channel/desc channelname = description
      channel/lock channelname = lockstring
      channel/unlock channelname = lockstring
      channel/ban channelname   (list bans)
      channel/ban[/quiet] channelname[, channelname, ...] = subscribername [: reason]
      channel/unban[/quiet] channelname[, channelname, ...] = subscribername
      channel/boot[/quiet] channelname[,channelname,...] = subscribername [: reason]

    # subtopics

    ## sending

    Usage: channel channelname msg
           channel channel name = msg  (with space in channel name)

    This sends a message to the channel. Note that you will rarely use this
    command like this; instead you can use the alias

        channelname <msg>
        channelalias <msg>

    For example

        public Hello World
        pub Hello World

    (this shortcut doesn't work for aliases containing spaces)

    See channel/alias for help on setting channel aliases.

    ## alias and unalias

    Usage: channel/alias channel = alias[;alias[;alias...]]
           channel/unalias alias
           channel    - this will list your subs and aliases to each channel

    Set one or more personal aliases for referencing a channel. For example:

        channel/alias warrior's guild = warrior;wguild;warchannel;warrior guild

    You can now send to the channel using all of these:

        warrior's guild Hello
        warrior Hello
        wguild Hello
        warchannel Hello

    Note that this will not work if the alias has a space in it. So the
    'warrior guild' alias must be used with the `channel` command:

        channel warrior guild = Hello

    Channel-aliases can be removed one at a time, using the '/unalias' switch.

    ## who

    Usage: channel/who channelname

    List the channel's subscribers. Shows who are currently offline or are
    muting the channel. Subscribers who are 'muting' will not see messages sent
    to the channel (use channel/mute to mute a channel).

    ## history

    Usage: channel/history channel [= index]

    This will display the last |c20|n lines of channel history. By supplying an
    index number, you will step that many lines back before viewing those 20 lines.

    For example:

        channel/history public = 35

    will go back 35 lines and show the previous 20 lines from that point (so
    lines -35 to -55).

    ## sub and unsub

    Usage: channel/sub channel [=alias[;alias;...]]
           channel/unsub channel

    This subscribes you to a channel and optionally assigns personal shortcuts
    for you to use to send to that channel (see aliases). When you unsub, all
    your personal aliases will also be removed.

    ## mute and unmute

    Usage: channel/mute channelname
           channel/unmute channelname

    Muting silences all output from the channel without actually
    un-subscribing. Other channel members will see that you are muted in the /who
    list. Sending a message to the channel will automatically unmute you.

    ## create and destroy

    Usage: channel/create channelname[;alias;alias[:typeclass]] [= description]
           channel/destroy channelname [= reason]

    Creates a new channel (or destroys one you control). You will automatically
    join the channel you create and everyone will be kicked and loose all aliases
    to a destroyed channel.

    ## lock and unlock

    Usage: channel/lock channelname = lockstring
           channel/unlock channelname = lockstring

    Note: this is an admin command.

    A lockstring is on the form locktype:lockfunc(). Channels understand three
    locktypes:
        listen - who may listen or join the channel.
        send - who may send messages to the channel
        control - who controls the channel. This is usually the one creating
            the channel.

    Common lockfuncs are all() and perm(). To make a channel everyone can
    listen to but only builders can talk on, use this:

        listen:all()
        send: perm(Builders)

    ## boot and ban

    Usage:
           channel/boot[/quiet] channelname[,channelname,...] = subscribername [: reason]
           channel/ban channelname[, channelname, ...] = subscribername [: reason]
           channel/unban channelname[, channelname, ...] = subscribername
           channel/unban channelname
           channel/ban channelname    (list bans)

    Booting will kick a named subscriber from channel(s) temporarily. The
    'reason' will be passed to the booted user. Unless the /quiet switch is
    used, the channel will also be informed of the action. A booted user is
    still able to re-connect, but they'll have to set up their aliases again.

    Banning will blacklist a user from (re)joining the provided channels. It
    will then proceed to boot them from those channels if they were connected.
    The 'reason' and `/quiet` works the same as for booting.

    Example:
        boot mychannel1 = EvilUser : Kicking you to cool down a bit.
        ban mychannel1,mychannel2= EvilUser : Was banned for spamming.

    """

    key = "@channel"
    aliases = ["@chan", "@channels"]
    help_category = "Comms"
    # these cmd: lock controls access to the channel command itself
    # the admin: lock controls access to /boot/ban/unban switches
    # the manage: lock  controls access to /create/destroy/desc/lock/unlock switches
    locks = "cmd:not pperm(channel_banned);admin:all();manage:all();changelocks:perm(Admin)"
    switch_options = (
        "list",
        "all",
        "history",
        "sub",
        "unsub",
        "mute",
        "unmute",
        "alias",
        "unalias",
        "create",
        "destroy",
        "desc",
        "lock",
        "unlock",
        "boot",
        "ban",
        "unban",
        "who",
    )
    # disable this in child command classes if wanting on-character channels
    account_caller = True

    def search_channel(self, channelname, exact=False, handle_errors=True):
        """
        Helper function for searching for a single channel with some error
        handling.

        Args:
            channelname (str): Name, alias #dbref or partial name/alias to search
                for.
            exact (bool, optional): If an exact or fuzzy-match of the name should be done.
                Note that even for a fuzzy match, an exactly given, unique channel name
                will always be returned.
            handle_errors (bool): If true, use `self.msg` to report errors if
                there are non/multiple matches. If so, the return will always be
                a single match or None.
        Returns:
            object, list or None: If `handle_errors` is `True`, this is either a found Channel
                or `None`. Otherwise it's a list  of zero, one or more channels found.
        Notes:
            The 'listen' and 'control' accesses are checked before returning.

        """
        caller = self.caller
        # first see if this is a personal alias
        channelname = caller.nicks.get(key=channelname, category="channel") or channelname

        # always try the exact match first.
        channels = CHANNEL_DEFAULT_TYPECLASS.objects.channel_search(channelname, exact=True)

        if not channels and not exact:
            # try fuzzy matching as well
            channels = CHANNEL_DEFAULT_TYPECLASS.objects.channel_search(channelname, exact=exact)

        # check permissions
        channels = [
            channel
            for channel in channels
            if channel.access(caller, "listen") or channel.access(caller, "control")
        ]

        if handle_errors:
            if not channels:
                self.msg(
                    f"No channel found matching '{channelname}' "
                    "(could also be due to missing access)."
                )
                return None
            elif len(channels) > 1:
                self.msg(
                    f"Multiple possible channel matches/alias for '{channelname}':\n"
                    + ", ".join(chan.key for chan in channels)
                )
                return None
            return channels[0]
        else:
            if not channels:
                return []
            elif len(channels) > 1:
                return list(channels)
            return [channels[0]]

    def msg_channel(self, channel, message, **kwargs):
        """
        Send a message to a given channel. This will check the 'send'
        permission on the channel.

        Args:
            channel (Channel): The channel to send to.
            message (str): The message to send.
            **kwargs: Unused by default. These kwargs will be passed into
                all channel messaging hooks for custom overriding.

        """
        if not channel.access(self.caller, "send"):
            self.msg(f"You are not allowed to send messages to channel {channel}")
            return

        # avoid unsafe tokens in message
        message = strip_unsafe_input(message, self.session)

        channel.msg(message, senders=self.caller, **kwargs)

    def get_channel_history(self, channel, start_index=0):
        """
        View a channel's history.

        Args:
            channel (Channel): The channel to access.
            message (str): The message to send.
            **kwargs: Unused by default. These kwargs will be passed into
                all channel messaging hooks for custom overriding.

        """
        caller = self.caller
        log_file = channel.get_log_filename()

        def send_msg(lines):
            return self.msg(
                "".join(line.split("[-]", 1)[1] if "[-]" in line else line for line in lines)
            )

        # asynchronously tail the log file
        tail_log_file(log_file, start_index, 20, callback=send_msg)

    def sub_to_channel(self, channel):
        """
        Subscribe to a channel. Note that all permissions should
        be checked before this step.

        Args:
            channel (Channel): The channel to access.

        Returns:
            bool, str: True, None if connection failed. If False,
                the second part is an error string.

        """
        caller = self.caller

        if channel.has_connection(caller):
            return False, f"Already listening to channel {channel.key}."

        # this sets up aliases in post_join_channel by default
        result = channel.connect(caller)

        return result, "" if result else f"Were not allowed to subscribe to channel {channel.key}"

    def unsub_from_channel(self, channel, **kwargs):
        """
        Un-Subscribe to a channel. Note that all permissions should
        be checked before this step.

        Args:
            channel (Channel): The channel to unsub from.
            **kwargs: Passed on to nick removal.

        Returns:
            bool, str: True, None if un-connection succeeded. If False,
                the second part is an error string.

        """
        caller = self.caller

        if not channel.has_connection(caller):
            return False, f"Not listening to channel {channel.key}."

        # this will also clean aliases
        result = channel.disconnect(caller)

        return result, "" if result else f"Could not unsubscribe from channel {channel.key}"

    def add_alias(self, channel, alias, **kwargs):
        """
        Add a new alias (nick) for the user to use with this channel.

        Args:
            channel (Channel): The channel to alias.
            alias (str): The personal alias to use for this channel.
            **kwargs: If given, passed into nicks.add.

        Note:
            We add two nicks - one is a plain `alias -> channel.key` that
            we need to be able to reference this channel easily. The other
            is a templated nick to easily be able to send messages to the
            channel without needing to give the full `channel` command. The
            structure of this nick is given by `self.channel_msg_pattern`
            and `self.channel_msg_nick_replacement`. By default it maps
            `alias <msg> -> channel <channelname> = <msg>`, so that you can
            for example just write `pub Hello` to send a message.

            The alias created is `alias $1 -> channel channel = $1`, to allow
            for sending to channel using the main channel command.

        """
        channel.add_user_channel_alias(self.caller, alias, **kwargs)

    def remove_alias(self, alias, **kwargs):
        """
        Remove an alias from a channel.

        Args:
            alias (str, optional): The alias to remove.
                The channel will be reverse-determined from the
                alias, if it exists.

        Returns:
            bool, str: True, None if removal succeeded. If False,
                the second part is an error string.
            **kwargs: If given, passed into nicks.get/add.

        Note:
            This will remove two nicks - the plain channel alias and the templated
            nick used for easily sending messages to the channel.

        """
        if self.caller.nicks.has(alias, category="channel", **kwargs):
            DefaultChannel.remove_user_channel_alias(self.caller, alias)
            return True, ""
        return False, "No such alias was defined."

    def get_channel_aliases(self, channel):
        """
        Get a user's aliases for a given channel. The user is retrieved
        through self.caller.

        Args:
            channel (Channel): The channel to act on.

        Returns:
            list: A list of zero, one or more alias-strings.

        """
        chan_key = channel.key.lower()
        nicktuples = self.caller.nicks.get(category="channel", return_tuple=True, return_list=True)
        if nicktuples:
            return [tup[2] for tup in nicktuples if tup[3].lower() == chan_key]
        return []

    def mute_channel(self, channel):
        """
        Temporarily mute a channel.

        Args:
            channel (Channel): The channel to alias.

        Returns:
            bool, str: True, None if muting successful. If False,
                the second part is an error string.
        """
        if channel.mute(self.caller):
            return True, ""
        return False, f"Channel {channel.key} was already muted."

    def unmute_channel(self, channel):
        """
        Unmute a channel.

        Args:
            channel (Channel): The channel to alias.

        Returns:
            bool, str: True, None if unmuting successful. If False,
                the second part is an error string.

        """
        if channel.unmute(self.caller):
            return True, ""
        return False, f"Channel {channel.key} was already unmuted."

    def create_channel(self, name, description, typeclass=None, aliases=None):
        """
        Create a new channel. Its name must not previously exist (case agnostic)
        (users can alias as needed). Will also connect to the new channel.

        Args:
            name (str): The new channel name/key.
            description (str): This is used in listings.
            aliases (list): A list of strings - alternative aliases for the channel
                (not to be confused with per-user aliases; these are available for
                everyone).

        Returns:
            channel, str: new_channel, "" if creation successful. If False,
                the second part is an error string.

        """
        caller = self.caller
        if typeclass:
            typeclass = class_from_module(typeclass)
        else:
            typeclass = CHANNEL_DEFAULT_TYPECLASS

        if typeclass.objects.channel_search(name, exact=True):
            return False, f"Channel {name} already exists."

        # set up the new channel
        lockstring = "send:all();listen:all();control:id(%s)" % caller.id

        new_chan = create.create_channel(
            name, aliases=aliases, desc=description, locks=lockstring, typeclass=typeclass
        )
        self.sub_to_channel(new_chan)
        return new_chan, ""

    def destroy_channel(self, channel, message=None):
        """
        Destroy an existing channel. Access should be checked before
        calling this function.

        Args:
            channel (Channel): The channel to alias.
            message (str, optional): Final message to send onto the channel
                before destroying it. If not given, a default message is
                used. Set to the empty string for no message.

        if typeclass:
           pass

        """
        caller = self.caller

        channel_key = channel.key
        if message is None:
            message = (
                f"|rChannel {channel_key} is being destroyed. "
                "Make sure to clean any channel aliases.|n"
            )
        if message:
            channel.msg(message, senders=caller, bypass_mute=True)
        channel.delete()
        logger.log_sec(f"Channel {channel_key} was deleted by {caller}")

    def set_lock(self, channel, lockstring):
        """
        Set a lockstring on a channel. Permissions must have been
        checked before this call.

        Args:
            channel (Channel): The channel to operate on.
            lockstring (str): A lockstring on the form 'type:lockfunc();...'

        Returns:
            bool, str: True, None if setting lock was successful. If False,
                the second part is an error string.

        """
        try:
            channel.locks.add(lockstring)
        except LockException as err:
            return False, err
        return True, ""

    def unset_lock(self, channel, lockstring):
        """
        Remove locks in a lockstring on a channel. Permissions must have been
        checked before this call.

        Args:
            channel (Channel): The channel to operate on.
            lockstring (str): A lockstring on the form 'type:lockfunc();...'

        Returns:
            bool, str: True, None if setting lock was successful. If False,
                the second part is an error string.

        """
        try:
            channel.locks.remove(lockstring)
        except LockException as err:
            return False, err
        return True, ""

    def set_desc(self, channel, description):
        """
        Set a channel description. This is shown in listings etc.

        Args:
            caller (Object or Account): The entity performing the action.
            channel (Channel): The channel to operate on.
            description (str): A short description of the channel.

        Returns:
            bool, str: True, None if setting lock was successful. If False,
                the second part is an error string.

        """
        channel.db.desc = description

    def boot_user(self, channel, target, quiet=False, reason=""):
        """
        Boot a user from a channel, with optional reason. This will
        also remove all their aliases for this channel.

        Args:
            channel (Channel): The channel to operate on.
            target (Object or Account): The entity to boot.
            quiet (bool, optional): Whether or not to announce to channel.
            reason (str, optional): A reason for the boot.

        Returns:
            bool, str: True, None if setting lock was successful. If False,
                the second part is an error string.

        """
        if not channel.subscriptions.has(target):
            return False, f"{target} is not connected to channel {channel.key}."
        # find all of target's nicks linked to this channel and delete them
        for nick in [
            nick
            for nick in target.nicks.get(category="channel", return_tuple=True) or []
            if nick.value[3].lower() == channel.key
        ]:
            nick.delete()
        channel.disconnect(target)
        reason = f" Reason: {reason}" if reason else ""
        target.msg(f"You were booted from channel {channel.key} by {self.caller.key}.{reason}")
        if not quiet:
            channel.msg(f"{target.key} was booted from channel by {self.caller.key}.{reason}")

        logger.log_sec(
            f"Channel Boot: {target} (Channel: {channel}, "
            f"Reason: {reason.strip()}, Caller: {self.caller}"
        )
        return True, ""

    def ban_user(self, channel, target, quiet=False, reason=""):
        """
        Ban a user from a channel, by locking them out. This will also
        boot them, if they are currently connected.

        Args:
            channel (Channel): The channel to operate on.
            target (Object or Account): The entity to ban
            quiet (bool, optional): Whether or not to announce to channel.
            reason (str, optional): A reason for the ban

        Returns:
            bool, str: True, None if banning was successful. If False,
                the second part is an error string.

        """
        self.boot_user(channel, target, quiet=quiet, reason=reason)
        if channel.ban(target):
            return True, ""
        return False, f"{target} is already banned from this channel."

    def unban_user(self, channel, target):
        """
        Un-Ban a user from a channel. This will not reconnect them
        to the channel, just allow them to connect again (assuming
        they have the suitable 'listen' lock like everyone else).

        Args:
            channel (Channel): The channel to operate on.
            target (Object or Account): The entity to unban

        Returns:
            bool, str: True, None if unbanning was successful. If False,
                the second part is an error string.

        """
        if channel.unban(target):
            return True, ""
        return False, f"{target} was not previously banned from this channel."

    def channel_list_bans(self, channel):
        """
        Show a channel's bans.

        Args:
            channel (Channel): The channel to operate on.

        Returns:
            list: A list of strings, each the name of a banned user.

        """
        return [banned.key for banned in channel.banlist]

    def channel_list_who(self, channel):
        """
        Show a list of online people is subscribing to a channel. This will check
        the 'control' permission of `caller` to determine if only online users
        should be returned or everyone.

        Args:
            channel (Channel): The channel to operate on.

        Returns:
            list: A list of prepared strings, with name + markers for if they are
                muted or offline.

        """
        caller = self.caller
        mute_list = list(channel.mutelist)
        online_list = channel.subscriptions.online()
        if channel.access(caller, "control"):
            # for those with channel control, show also offline users
            all_subs = list(channel.subscriptions.all())
        else:
            # for others, only show online users
            all_subs = online_list

        who_list = []
        for subscriber in all_subs:
            name = subscriber.get_display_name(caller)
            conditions = (
                "muting" if subscriber in mute_list else "",
                "offline" if subscriber not in online_list else "",
            )
            conditions = [cond for cond in conditions if cond]
            cond_text = "(" + ", ".join(conditions) + ")" if conditions else ""
            who_list.append(f"{name}{cond_text}")

        return who_list

    def list_channels(self, channelcls=CHANNEL_DEFAULT_TYPECLASS):
        """
        Return a available channels.

        Args:
            channelcls (Channel, optional): The channel-class to query on. Defaults
                to the default channel class from settings.

        Returns:
            tuple: A tuple `(subbed_chans, available_chans)` with the channels
                currently subscribed to, and those we have 'listen' access to but
                don't actually sub to yet.

        """
        caller = self.caller
        subscribed_channels = list(channelcls.objects.get_subscriptions(caller))
        unsubscribed_available_channels = [
            chan
            for chan in channelcls.objects.get_all_channels()
            if chan not in subscribed_channels and chan.access(caller, "listen")
        ]
        return subscribed_channels, unsubscribed_available_channels

    def display_subbed_channels(self, subscribed):
        """
        Display channels subscribed to.

        Args:
            subscribed (list): List of subscribed channels

        Returns:
            EvTable: Table to display.

        """
        comtable = self.styled_table(
            "id",
            "channel",
            "my aliases",
            "locks",
            "description",
            align="l",
            maxwidth=_DEFAULT_WIDTH,
        )
        for chan in subscribed:
            locks = "-"
            chanid = "-"
            if chan.access(self.caller, "control"):
                locks = chan.locks
                chanid = chan.id

            my_aliases = ", ".join(self.get_channel_aliases(chan))
            comtable.add_row(
                *(
                    chanid,
                    "{key}{aliases}".format(
                        key=chan.key,
                        aliases=";" + ";".join(chan.aliases.all()) if chan.aliases.all() else "",
                    ),
                    my_aliases,
                    locks,
                    chan.db.desc,
                )
            )
        return comtable

    def display_all_channels(self, subscribed, available):
        """
        Display all available channels

        Args:
            subscribed (list): List of subscribed channels
        Returns:
            EvTable: Table to display.

        """
        caller = self.caller

        comtable = self.styled_table(
            "sub",
            "channel",
            "aliases",
            "my aliases",
            "description",
            maxwidth=_DEFAULT_WIDTH,
        )
        channels = subscribed + available

        for chan in channels:
            if chan not in subscribed:
                substatus = "|rNo|n"
            elif caller in chan.mutelist:
                substatus = "|rMuting|n"
            else:
                substatus = "|gYes|n"
            my_aliases = ", ".join(self.get_channel_aliases(chan))
            comtable.add_row(
                *(
                    substatus,
                    chan.key,
                    ",".join(chan.aliases.all()) if chan.aliases.all() else "",
                    my_aliases,
                    chan.db.desc,
                )
            )
        comtable.reformat_column(0, width=8)

        return comtable

    def func(self):
        """
        Main functionality of command.
        """
        # from evennia import set_trace;set_trace()

        caller = self.caller
        switches = self.switches
        channel_names = [name for name in self.lhslist if name]

        # from evennia import set_trace;set_trace()

        if "all" in switches:
            # show all available channels
            subscribed, available = self.list_channels()
            table = self.display_all_channels(subscribed, available)

            self.msg(
                "\n|wAvailable channels|n (use no argument to "
                f"only show your subscriptions)\n{table}"
            )
            return

        if not channel_names:
            # empty arg show only subscribed channels
            subscribed, _ = self.list_channels()
            table = self.display_subbed_channels(subscribed)

            self.msg(f"\n|wChannel subscriptions|n (use |w/all|n to see all available):\n{table}")
            return

        if not self.switches and not self.args:
            self.msg("Usage[/switches]: channel [= message]")
            return

        if "create" in switches:
            # create a new channel
            if not self.access(caller, "manage"):
                self.msg("You don't have access to use channel/create.")
                return

            config = self.lhs
            if not config:
                self.msg("To create: channel/create name[;aliases][:typeclass] [= description]")
                return
            name, *typeclass = config.rsplit(":", 1)
            typeclass = typeclass[0] if typeclass else None
            name, *aliases = name.rsplit(";")
            description = self.rhs or ""
            chan, err = self.create_channel(name, description, typeclass=typeclass, aliases=aliases)
            if chan:
                self.msg(f"Created (and joined) new channel '{chan.key}'.")
            else:
                self.msg(err)
            return

        if "unalias" in switches:
            # remove a personal alias (no channel needed)
            alias = self.args.strip()
            if not alias:
                self.msg("Specify the alias to remove as channel/unalias <alias>")
                return
            success, err = self.remove_alias(alias)
            if success:
                self.msg(f"Removed your channel alias '{alias}'.")
            else:
                self.msg(err)
            return

        possible_lhs_message = ""
        if not self.rhs and self.args and " " in self.args:
            # since we want to support messaging with `channel name text` (for
            # channels without a space in their name), we need to check if the
            # first 'channel name' is in fact 'channelname text'
            no_rhs_channel_name = self.args.split(" ", 1)[0]
            possible_lhs_message = self.args[len(no_rhs_channel_name) :]
            if possible_lhs_message.strip() == "=":
                possible_lhs_message = ""
            channel_names.append(no_rhs_channel_name)

        channels = []
        errors = []
        for channel_name in channel_names:
            # find a channel by fuzzy-matching. This also checks
            # 'listen/control' perms.
            found_channels = self.search_channel(channel_name, exact=False, handle_errors=False)
            if not found_channels:
                errors.append(
                    f"No channel found matching '{channel_name}' "
                    "(could also be due to missing access)."
                )
            elif len(found_channels) > 1:
                errors.append(
                    f"Multiple possible channel matches/alias for '{channel_name}':\n"
                    + ", ".join(chan.key for chan in found_channels)
                )
            else:
                channels.append(found_channels[0])

        if not channels:
            self.msg("\n".join(errors))
            return

        # we have at least one channel at this point
        channel = channels[0]

        if not switches:
            if self.rhs:
                # send message to channel
                self.msg_channel(channel, self.rhs.strip())
            elif channel and possible_lhs_message:
                # called on the form channelname message without =
                self.msg_channel(channel, possible_lhs_message.strip())
            else:
                # inspect a given channel
                subscribed, available = self.list_channels()
                if channel in subscribed:
                    table = self.display_subbed_channels([channel])
                    header = f"Channel |w{channel.key}|n"
                    self.msg(
                        f"{header}\n(use |w{channel.key} <msg>|n (or a channel-alias) "
                        "to chat and the 'channel' command "
                        f"to customize)\n{table}"
                    )
                elif channel in available:
                    table = self.display_all_channels([], [channel])
                    self.msg(
                        "\n|wNot subscribed to this channel|n (use /list to "
                        f"show all subscriptions)\n{table}"
                    )
            return

        if "history" in switches or "hist" in switches:
            # view channel history

            index = self.rhs or 0
            try:
                index = max(0, int(index))
            except ValueError:
                self.msg(
                    "The history index (describing how many lines to go back) "
                    "must be an integer >= 0."
                )
                return
            self.get_channel_history(channel, start_index=index)
            return

        if "sub" in switches:
            # subscribe to a channel
            aliases = []
            if self.rhs:
                aliases = set(alias.strip().lower() for alias in self.rhs.split(";"))
            success, err = self.sub_to_channel(channel)
            if success:
                for alias in aliases:
                    self.add_alias(channel, alias)
                alias_txt = ", ".join(aliases)
                alias_txt = f" using alias(es) {alias_txt}" if aliases else ""
                self.msg(
                    "You are now subscribed "
                    f"to the channel {channel.key}{alias_txt}. Use /alias to "
                    "add additional aliases for referring to the channel."
                )
            else:
                self.msg(err)
            return

        if "unsub" in switches:
            # un-subscribe from a channel
            success, err = self.unsub_from_channel(channel)
            if success:
                self.msg(f"You un-subscribed from channel {channel.key}. All aliases were cleared.")
            else:
                self.msg(err)
            return

        if "alias" in switches:
            # create a new personal alias for a channel
            alias = self.rhs
            if not alias:
                self.msg("Specify the alias as channel/alias channelname = alias")
                return
            self.add_alias(channel, alias)
            self.msg(f"Added/updated your alias '{alias}' for channel {channel.key}.")
            return

        if "mute" in switches:
            # mute a given channel
            success, err = self.mute_channel(channel)
            if success:
                self.msg(f"Muted channel {channel.key}.")
            else:
                self.msg(err)
            return

        if "unmute" in switches:
            # unmute a given channel
            success, err = self.unmute_channel(channel)
            if success:
                self.msg(f"Un-muted channel {channel.key}.")
            else:
                self.msg(err)
            return

        if "destroy" in switches or "delete" in switches:
            # destroy a channel we control

            if not self.access(caller, "manage"):
                self.msg("You don't have access to use channel/destroy.")
                return

            if not channel.access(caller, "control"):
                self.msg("You can only delete channels you control.")
                return

            reason = self.rhs or None

            def _perform_delete(caller, *args, **kwargs):
                self.destroy_channel(channel, message=reason)
                self.msg(f"Channel {channel.key} was successfully deleted.")

            ask_yes_no(
                caller,
                prompt=(
                    f"Are you sure you want to delete channel '{channel.key}' "
                    "(make sure name is correct!)?\nThis will disconnect and "
                    "remove all users' aliases. {options}?"
                ),
                yes_action=_perform_delete,
                no_action="Aborted.",
                default="N",
            )

        if "desc" in switches:
            # set channel description

            if not self.access(caller, "manage"):
                self.msg("You don't have access to use channel/desc.")
                return

            if not channel.access(caller, "control"):
                self.msg("You can only change description of channels you control.")
                return

            desc = self.rhs.strip()

            if not desc:
                self.msg("Usage: /desc channel = description")
                return

            self.set_desc(channel, desc)
            self.msg("Updated channel description.")

        if "lock" in switches:
            # add a lockstring to channel

            if not self.access(caller, "changelocks"):
                self.msg("You don't have access to use channel/lock.")
                return

            if not channel.access(caller, "control"):
                self.msg("You need 'control'-access to change locks on this channel.")
                return

            lockstring = self.rhs.strip()

            if not lockstring:
                self.msg("Usage: channel/lock channelname = lockstring")
                return

            success, err = self.set_lock(channel, self.rhs)
            if success:
                self.msg("Added/updated lock on channel.")
            else:
                self.msg(f"Could not add/update lock: {err}")
            return

        if "unlock" in switches:
            # remove/update lockstring from channel

            if not self.access(caller, "changelocks"):
                self.msg("You don't have access to use channel/unlock.")
                return

            if not channel.access(caller, "control"):
                self.msg("You need 'control'-access to change locks on this channel.")
                return

            lockstring = self.rhs.strip()

            if not lockstring:
                self.msg("Usage: channel/unlock channelname = lockstring")
                return

            success, err = self.unset_lock(channel, self.rhs)
            if success:
                self.msg("Removed lock from channel.")
            else:
                self.msg(f"Could not remove lock: {err}")
            return

        if "boot" in switches:
            # boot a user from channel(s)

            if not self.access(caller, "admin"):
                self.msg("You don't have access to use channel/boot.")
                return

            if not self.rhs:
                self.msg("Usage: channel/boot channel[,channel,...] = username [:reason]")
                return

            target_str, *reason = self.rhs.rsplit(":", 1)
            reason = reason[0].strip() if reason else ""

            for chan in channels:
                if not chan.access(caller, "control"):
                    self.msg(f"You need 'control'-access to boot a user from {chan.key}.")
                    return

                # the target must be a member of all given channels
                target = caller.search(target_str, candidates=chan.subscriptions.all())
                if not target:
                    self.msg(f"Cannot boot '{target_str}' - not in channel {chan.key}.")
                    return

            def _boot_user(caller, *args, **kwargs):
                for chan in channels:
                    success, err = self.boot_user(chan, target, quiet=False, reason=reason)
                    if success:
                        self.msg(f"Booted {target.key} from channel {chan.key}.")
                    else:
                        self.msg(f"Cannot boot {target.key} from channel {chan.key}: {err}")

            channames = ", ".join(chan.key for chan in channels)
            reasonwarn = (
                ". Also note that your reason will be echoed to the channel" if reason else ""
            )
            ask_yes_no(
                caller,
                prompt=(
                    f"Are you sure you want to boot user {target.key} from "
                    f"channel(s) {channames} (make sure name/channels are correct{reasonwarn}). "
                    "{options}?"
                ),
                yes_action=_boot_user,
                no_action="Aborted.",
                default="Y",
            )
            return

        if "ban" in switches:
            # ban a user from channel(s)

            if not self.access(caller, "admin"):
                self.msg("You don't have access to use channel/ban.")
                return

            if not self.rhs:
                # view bans for channels

                if not channel.access(caller, "control"):
                    self.msg(f"You need 'control'-access to view bans on channel {channel.key}")
                    return

                bans = [
                    "Channel bans "
                    "(to ban, use channel/ban channel[,channel,...] = username [:reason]"
                ]
                bans.extend(self.channel_list_bans(channel))
                self.msg("\n".join(bans))
                return

            target_str, *reason = self.rhs.rsplit(":", 1)
            reason = reason[0].strip() if reason else ""

            for chan in channels:
                # the target must be a member of all given channels
                if not chan.access(caller, "control"):
                    self.msg(f"You don't have access to ban users on channel {chan.key}")
                    return

                target = caller.search(target_str, candidates=chan.subscriptions.all())

                if not target:
                    self.msg(f"Cannot ban '{target_str}' - not in channel {chan.key}.")
                    return

            def _ban_user(caller, *args, **kwargs):
                for chan in channels:
                    success, err = self.ban_user(chan, target, quiet=False, reason=reason)
                    if success:
                        self.msg(f"Banned {target.key} from channel {chan.key}.")
                    else:
                        self.msg(f"Cannot boot {target.key} from channel {chan.key}: {err}")

            channames = ", ".join(chan.key for chan in channels)
            reasonwarn = (
                ". Also note that your reason will be echoed to the channel" if reason else ""
            )
            ask_yes_no(
                caller,
                (
                    f"Are you sure you want to ban user {target.key} from "
                    f"channel(s) {channames} (make sure name/channels are correct{reasonwarn}) "
                    "{options}?"
                ),
                _ban_user,
                "Aborted.",
            )
            return

        if "unban" in switches:
            # unban a previously banned user from channel

            if not self.access(caller, "admin"):
                self.msg("You don't have access to use channel/unban.")
                return

            target_str = self.rhs.strip()

            if not target_str:
                self.msg("Usage: channel[,channel,...] = user")
                return

            banlists = []
            for chan in channels:
                # the target must be a member of all given channels
                if not chan.access(caller, "control"):
                    self.msg(f"You don't have access to unban users on channel {chan.key}")
                    return
                banlists.extend(chan.banlist)

            target = caller.search(target_str, candidates=banlists)
            if not target:
                self.msg(f"Could not find a banned user '{target_str}' in given channel(s).")
                return

            for chan in channels:
                success, err = self.unban_user(channel, target)
                if success:
                    self.msg(f"Un-banned {target_str} from channel {chan.key}")
                else:
                    self.msg(err)
            return

        if "who" in switches:
            # view who's a member of a channel

            who_list = [f"Subscribed to {channel.key}:"]
            who_list.extend(self.channel_list_who(channel))
            self.msg("\n".join(who_list))
            return


# a channel-command parent for use with Characters/Objects.
class CmdObjectChannel(CmdChannel):
    account_caller = False


class CmdPage(COMMAND_DEFAULT_CLASS):
    """
    send a private message to another account

    Usage:
      page <account> <message>
      page[/switches] [<account>,<account>,... = <message>]
      tell        ''
      page <number>

    Switches:
      last - shows who you last messaged
      list - show your last <number> of tells/pages (default)

    Send a message to target user (if online). If no argument is given, you
    will get a list of your latest messages. The equal sign is needed for
    multiple targets or if sending to target with space in the name.

    """

    key = "page"
    aliases = ["tell"]
    switch_options = ("last", "list")
    locks = "cmd:not pperm(page_banned)"
    help_category = "Comms"

    # this is used by the COMMAND_DEFAULT_CLASS parent
    account_caller = True

    def func(self):
        """Implement function using the Msg methods"""

        # Since account_caller is set above, this will be an Account.
        caller = self.caller

        # get the messages we've sent (not to channels)
        pages_we_sent = Msg.objects.get_messages_by_sender(caller).order_by("-db_date_created")
        # get last messages we've got
        pages_we_got = Msg.objects.get_messages_by_receiver(caller).order_by("-db_date_created")
        targets, message, number = [], None, None

        if "last" in self.switches:
            if pages_we_sent:
                recv = ",".join(obj.key for obj in pages_we_sent[0].receivers)
                self.msg(f"You last paged |c{recv}|n:{pages_we_sent[0].message}")
                return
            else:
                self.msg("You haven't paged anyone yet.")
                return

        if self.args:
            if self.rhs:
                for target in self.lhslist:
                    target_obj = self.caller.search(target)
                    if not target_obj:
                        return
                    targets.append(target_obj)
                message = self.rhs.strip()
            else:
                target, *message = self.args.split(" ", 1)
                if target and target.isnumeric():
                    # a number to specify a historic page
                    number = int(target)
                elif target:
                    target_obj = self.caller.search(target, quiet=True)
                    if target_obj:
                        # a proper target
                        targets = [target_obj[0]]
                        message = message[0].strip()
                    else:
                        # a message with a space in it - put it back together
                        message = target + " " + (message[0] if message else "")
                else:
                    # a single-word message
                    message = message[0].strip()

        pages = list(pages_we_sent) + list(pages_we_got)
        pages = sorted(pages, key=lambda page: page.date_created)

        if message:
            # send a message
            if not targets:
                # no target given - send to last person we paged
                if pages_we_sent:
                    targets = pages_we_sent[0].receivers
                else:
                    self.msg("Who do you want page?")
                    return

            header = f"|wAccount|n |c{caller.key}|n |wpages:|n"
            if message.startswith(":"):
                message = f"{caller.key} {message.strip(':').strip()}"

            # create the persistent message object
            create.create_message(caller, message, receivers=targets)

            # tell the accounts they got a message.
            received = []
            rstrings = []
            for target in targets:
                if not target.access(caller, "msg"):
                    rstrings.append(f"You are not allowed to page {target}.")
                    continue
                target.msg(f"{header} {message}")
                if hasattr(target, "sessions") and not target.sessions.count():
                    received.append(f"|C{target.name}|n")
                    rstrings.append(
                        f"{received[-1]} is offline. They will see your message "
                        "if they list their pages later."
                    )
                else:
                    received.append(f"|c{target.name}|n")
            if rstrings:
                self.msg("\n".join(rstrings))
            self.msg("You paged %s with: '%s'." % (", ".join(received), message))
            return

        else:
            # no message to send
            if number is not None and len(pages) > number:
                lastpages = pages[-number:]
            else:
                lastpages = pages
            to_template = "|w{date}{clr} {sender}|nto{clr}{receiver}|n:> {message}"
            from_template = "|w{date}{clr} {receiver}|nfrom{clr}{sender}|n:< {message}"
            listing = []
            prev_selfsend = False
            for page in lastpages:
                multi_send = len(page.senders) > 1
                multi_recv = len(page.receivers) > 1
                sending = self.caller in page.senders
                # self-messages all look like sends, so we assume they always
                # come in close pairs and treat the second of the pair as the recv.
                selfsend = sending and self.caller in page.receivers
                if selfsend:
                    if prev_selfsend:
                        # this is actually a receive of a self-message
                        sending = False
                        prev_selfsend = False
                    else:
                        prev_selfsend = True

                clr = "|c" if sending else "|g"

                sender = f"|n,{clr}".join(obj.key for obj in page.senders)
                receiver = f"|n,{clr}".join([obj.name for obj in page.receivers])
                if sending:
                    template = to_template
                    sender = f"{sender} " if multi_send else ""
                    receiver = f" {receiver}" if multi_recv else f" {receiver}"
                else:
                    template = from_template
                    receiver = f"{receiver} " if multi_recv else ""
                    sender = f" {sender} " if multi_send else f" {sender}"

                listing.append(
                    template.format(
                        date=utils.datetime_format(page.date_created),
                        clr=clr,
                        sender=sender,
                        receiver=receiver,
                        message=page.message,
                    )
                )
            lastpages = "\n ".join(listing)

            if lastpages:
                string = f"Your latest pages:\n {lastpages}"
            else:
                string = "You haven't paged anyone yet."
            self.msg(string)
            return


def _list_bots(cmd):
    """
    Helper function to produce a list of all IRC bots.

    Args:
        cmd (Command): Instance of the Bot command.
    Returns:
        bots (str): A table of bots or an error message.

    """
    ircbots = [
        bot for bot in AccountDB.objects.filter(db_is_bot=True, username__startswith="ircbot-")
    ]
    if ircbots:
        table = cmd.styled_table(
            "|w#dbref|n",
            "|wbotname|n",
            "|wev-channel|n",
            "|wirc-channel|n",
            "|wSSL|n",
            maxwidth=_DEFAULT_WIDTH,
        )
        for ircbot in ircbots:
            ircinfo = "%s (%s:%s)" % (
                ircbot.db.irc_channel,
                ircbot.db.irc_network,
                ircbot.db.irc_port,
            )
            table.add_row(
                "#%i" % ircbot.id,
                ircbot.db.irc_botname,
                ircbot.db.ev_channel,
                ircinfo,
                ircbot.db.irc_ssl,
            )
        return table
    else:
        return "No irc bots found."


class CmdIRC2Chan(COMMAND_DEFAULT_CLASS):
    """
    Link an evennia channel to an external IRC channel

    Usage:
      irc2chan[/switches] <evennia_channel> = <ircnetwork> <port> <#irchannel> <botname>[:typeclass]
      irc2chan/delete botname|#dbid

    Switches:
      /delete     - this will delete the bot and remove the irc connection
                    to the channel. Requires the botname or #dbid as input.
      /remove     - alias to /delete
      /disconnect - alias to /delete
      /list       - show all irc<->evennia mappings
      /ssl        - use an SSL-encrypted connection

    Example:
      irc2chan myircchan = irc.dalnet.net 6667 #mychannel evennia-bot
      irc2chan public = irc.freenode.net 6667 #evgaming #evbot:accounts.mybot.MyBot

    This creates an IRC bot that connects to a given IRC network and
    channel. If a custom typeclass path is given, this will be used
    instead of the default bot class.
    The bot will relay everything said in the evennia channel to the
    IRC channel and vice versa. The bot will automatically connect at
    server start, so this command need only be given once. The
    /disconnect switch will permanently delete the bot. To only
    temporarily deactivate it, use the  |wservices|n command instead.
    Provide an optional bot class path to use a custom bot.
    """

    key = "irc2chan"
    switch_options = ("delete", "remove", "disconnect", "list", "ssl")
    locks = "cmd:serversetting(IRC_ENABLED) and pperm(Developer)"
    help_category = "Comms"

    def func(self):
        """Setup the irc-channel mapping"""

        if not settings.IRC_ENABLED:
            string = """IRC is not enabled. You need to activate it in game/settings.py."""
            self.msg(string)
            return

        if "list" in self.switches:
            # show all connections
            self.msg(_list_bots(self))
            return

        if "disconnect" in self.switches or "remove" in self.switches or "delete" in self.switches:
            botname = f"ircbot-{self.lhs}"
            matches = AccountDB.objects.filter(db_is_bot=True, username=botname)
            dbref = utils.dbref(self.lhs)
            if not matches and dbref:
                # try dbref match
                matches = AccountDB.objects.filter(db_is_bot=True, id=dbref)
            if matches:
                matches[0].delete()
                self.msg("IRC connection destroyed.")
            else:
                self.msg("IRC connection/bot could not be removed, does it exist?")
            return

        if not self.args or not self.rhs:
            string = (
                "Usage: irc2chan[/switches] <evennia_channel> ="
                " <ircnetwork> <port> <#irchannel> <botname>[:typeclass]"
            )
            self.msg(string)
            return

        channel = self.lhs
        self.rhs = self.rhs.replace("#", " ")  # to avoid Python comment issues
        try:
            irc_network, irc_port, irc_channel, irc_botname = [
                part.strip() for part in self.rhs.split(None, 4)
            ]
            irc_channel = f"#{irc_channel}"
        except Exception:
            string = "IRC bot definition '%s' is not valid." % self.rhs
            self.msg(string)
            return

        botclass = None
        if ":" in irc_botname:
            irc_botname, botclass = [part.strip() for part in irc_botname.split(":", 2)]
        botname = f"ircbot-{irc_botname}"
        # If path given, use custom bot otherwise use default.
        botclass = botclass if botclass else bots.IRCBot
        irc_ssl = "ssl" in self.switches

        # create a new bot
        bot = AccountDB.objects.filter(username__iexact=botname)
        if bot:
            # re-use an existing bot
            bot = bot[0]
            if not bot.is_bot:
                self.msg(f"Account '{botname}' already exists and is not a bot.")
                return
        else:
            try:
                bot = create.create_account(botname, None, None, typeclass=botclass)
            except Exception as err:
                self.msg(f"|rError, could not create the bot:|n '{err}'.")
                return
        bot.start(
            ev_channel=channel,
            irc_botname=irc_botname,
            irc_channel=irc_channel,
            irc_network=irc_network,
            irc_port=irc_port,
            irc_ssl=irc_ssl,
        )
        self.msg("Connection created. Starting IRC bot.")


class CmdIRCStatus(COMMAND_DEFAULT_CLASS):
    """
    Check and reboot IRC bot.

    Usage:
        ircstatus [#dbref ping | nicklist | reconnect]

    If not given arguments, will return a list of all bots (like
    irc2chan/list). The 'ping' argument will ping the IRC network to
    see if the connection is still responsive. The 'nicklist' argument
    (aliases are 'who' and 'users') will return a list of users on the
    remote IRC channel.  Finally, 'reconnect' will force the client to
    disconnect and reconnect again. This may be a last resort if the
    client has silently lost connection (this may happen if the remote
    network experience network issues). During the reconnection
    messages sent to either channel will be lost.

    """

    key = "ircstatus"
    locks = "cmd:serversetting(IRC_ENABLED) and perm(ircstatus) or perm(Builder))"
    help_category = "Comms"

    def func(self):
        """Handles the functioning of the command."""

        if not self.args:
            self.msg(_list_bots(self))
            return
        # should always be on the form botname option
        args = self.args.split()
        if len(args) != 2:
            self.msg("Usage: ircstatus [#dbref ping||nicklist||reconnect]")
            return
        botname, option = args
        if option not in ("ping", "users", "reconnect", "nicklist", "who"):
            self.msg("Not a valid option.")
            return
        matches = None
        if utils.dbref(botname):
            matches = AccountDB.objects.filter(db_is_bot=True, id=utils.dbref(botname))
        if not matches:
            self.msg(
                "No matching IRC-bot found. Use ircstatus without arguments to list active bots."
            )
            return
        ircbot = matches[0]
        channel = ircbot.db.irc_channel
        network = ircbot.db.irc_network
        port = ircbot.db.irc_port
        chtext = f"IRC bot '{ircbot.db.irc_botname}' on channel {channel} ({network}:{port})"
        if option == "ping":
            # check connection by sending outself a ping through the server.
            self.msg(f"Pinging through {chtext}.")
            ircbot.ping(self.caller)
        elif option in ("users", "nicklist", "who"):
            # retrieve user list. The bot must handles the echo since it's
            # an asynchronous call.
            self.msg(f"Requesting nicklist from {channel} ({network}:{port}).")
            ircbot.get_nicklist(self.caller)
        elif self.caller.locks.check_lockstring(
            self.caller, "dummy:perm(ircstatus) or perm(Developer)"
        ):
            # reboot the client
            self.msg(f"Forcing a disconnect + reconnect of {chtext}.")
            ircbot.reconnect()
        else:
            self.msg("You don't have permission to force-reload the IRC bot.")


# RSS connection
class CmdRSS2Chan(COMMAND_DEFAULT_CLASS):
    """
    link an evennia channel to an external RSS feed

    Usage:
      rss2chan[/switches] <evennia_channel> = <rss_url>

    Switches:
      /disconnect - this will stop the feed and remove the connection to the
                    channel.
      /remove     -                                 "
      /list       - show all rss->evennia mappings

    Example:
      rss2chan rsschan = http://code.google.com/feeds/p/evennia/updates/basic

    This creates an RSS reader  that connects to a given RSS feed url. Updates
    will be echoed as a title and news link to the given channel. The rate of
    updating is set with the RSS_UPDATE_INTERVAL variable in settings (default
    is every 10 minutes).

    When disconnecting you need to supply both the channel and url again so as
    to identify the connection uniquely.
    """

    key = "rss2chan"
    switch_options = ("disconnect", "remove", "list")
    locks = "cmd:serversetting(RSS_ENABLED) and pperm(Developer)"
    help_category = "Comms"

    def func(self):
        """Setup the rss-channel mapping"""

        # checking we have all we need
        if not settings.RSS_ENABLED:
            string = """RSS is not enabled. You need to activate it in game/settings.py."""
            self.msg(string)
            return
        try:
            import feedparser

            assert feedparser  # to avoid checker error of not being used
        except ImportError:
            string = (
                "RSS requires python-feedparser (https://pypi.python.org/pypi/feedparser)."
                " Install before continuing."
            )
            self.msg(string)
            return

        if "list" in self.switches:
            # show all connections
            rssbots = [
                bot
                for bot in AccountDB.objects.filter(db_is_bot=True, username__startswith="rssbot-")
            ]
            if rssbots:
                table = self.styled_table(
                    "|wdbid|n",
                    "|wupdate rate|n",
                    "|wev-channel",
                    "|wRSS feed URL|n",
                    border="cells",
                    maxwidth=_DEFAULT_WIDTH,
                )
                for rssbot in rssbots:
                    table.add_row(
                        rssbot.id, rssbot.db.rss_rate, rssbot.db.ev_channel, rssbot.db.rss_url
                    )
                self.msg(table)
            else:
                self.msg("No rss bots found.")
            return

        if "disconnect" in self.switches or "remove" in self.switches or "delete" in self.switches:
            botname = f"rssbot-{self.lhs}"
            matches = AccountDB.objects.filter(db_is_bot=True, db_key=botname)
            if not matches:
                # try dbref match
                matches = AccountDB.objects.filter(db_is_bot=True, id=self.args.lstrip("#"))
            if matches:
                matches[0].delete()
                self.msg("RSS connection destroyed.")
            else:
                self.msg("RSS connection/bot could not be removed, does it exist?")
            return

        if not self.args or not self.rhs:
            string = "Usage: rss2chan[/switches] <evennia_channel> = <rss url>"
            self.msg(string)
            return
        channel = self.lhs
        url = self.rhs

        botname = f"rssbot-{url}"
        bot = AccountDB.objects.filter(username__iexact=botname)
        if bot:
            # re-use existing bot
            bot = bot[0]
            if not bot.is_bot:
                self.msg(f"Account '{botname}' already exists and is not a bot.")
                return
        else:
            # create a new bot
            bot = create.create_account(botname, None, None, typeclass=bots.RSSBot)
        bot.start(ev_channel=channel, rss_url=url, rss_rate=10)
        self.msg("RSS reporter created. Fetching RSS.")


class CmdGrapevine2Chan(COMMAND_DEFAULT_CLASS):
    """
    Link an Evennia channel to an external Grapevine channel

    Usage:
      grapevine2chan[/switches] <evennia_channel> = <grapevine_channel>
      grapevine2chan/disconnect <connection #id>

    Switches:
        /list     - (or no switch): show existing grapevine <-> Evennia
                    mappings and available grapevine chans
        /remove   - alias to disconnect
        /delete   - alias to disconnect

    Example:
        grapevine2chan mygrapevine = gossip

    This creates a link between an in-game Evennia channel and an external
    Grapevine channel. The game must be registered with the Grapevine network
    (register at https://grapevine.haus) and the GRAPEVINE_* auth information
    must be added to game settings.
    """

    key = "grapevine2chan"
    switch_options = ("disconnect", "remove", "delete", "list")
    locks = "cmd:serversetting(GRAPEVINE_ENABLED) and pperm(Developer)"
    help_category = "Comms"

    def func(self):
        """Setup the Grapevine channel mapping"""

        if not settings.GRAPEVINE_ENABLED:
            self.msg("Set GRAPEVINE_ENABLED=True in settings to enable.")
            return

        if "list" in self.switches:
            # show all connections
            gwbots = [
                bot
                for bot in AccountDB.objects.filter(
                    db_is_bot=True, username__startswith="grapevinebot-"
                )
            ]
            if gwbots:
                table = self.styled_table(
                    "|wdbid|n",
                    "|wev-channel",
                    "|wgw-channel|n",
                    border="cells",
                    maxwidth=_DEFAULT_WIDTH,
                )
                for gwbot in gwbots:
                    table.add_row(gwbot.id, gwbot.db.ev_channel, gwbot.db.grapevine_channel)
                self.msg(table)
            else:
                self.msg("No grapevine bots found.")
            return

        if "disconnect" in self.switches or "remove" in self.switches or "delete" in self.switches:
            botname = f"grapevinebot-{self.lhs}"
            matches = AccountDB.objects.filter(db_is_bot=True, db_key=botname)

            if not matches:
                # try dbref match
                matches = AccountDB.objects.filter(db_is_bot=True, id=self.args.lstrip("#"))
            if matches:
                matches[0].delete()
                self.msg("Grapevine connection destroyed.")
            else:
                self.msg("Grapevine connection/bot could not be removed, does it exist?")
            return

        if not self.args or not self.rhs:
            string = "Usage: grapevine2chan[/switches] <evennia_channel> = <grapevine_channel>"
            self.msg(string)
            return

        channel = self.lhs
        grapevine_channel = self.rhs

        botname = "grapewinebot-%s-%s" % (channel, grapevine_channel)
        bot = AccountDB.objects.filter(username__iexact=botname)
        if bot:
            # re-use existing bot
            bot = bot[0]
            if not bot.is_bot:
                self.msg(f"Account '{botname}' already exists and is not a bot.")
                return
            else:
                self.msg(f"Reusing bot '{botname}' ({bot.dbref})")
        else:
            # create a new bot
            bot = create.create_account(botname, None, None, typeclass=bots.GrapevineBot)

        bot.start(ev_channel=channel, grapevine_channel=grapevine_channel)
        self.msg(f"Grapevine connection created {channel} <-> {grapevine_channel}.")


class CmdDiscord2Chan(COMMAND_DEFAULT_CLASS):
    """
    Link an Evennia channel to an external Discord channel

    Usage:
      discord2chan[/switches]
      discord2chan[/switches] <evennia_channel> [= <discord_channel_id>]

    Switches:
        /list    - (or no switch) show existing Evennia <-> Discord links
        /remove  - remove an existing link by link ID
        /delete  - alias to remove
        /guild   - toggle the Discord server tag on/off
        /channel - toggle the Evennia/Discord channel tags on/off
        /start   - tell the bot to start, in case it lost its connection

    Example:
        discord2chan mydiscord = 555555555555555

    This creates a link between an in-game Evennia channel and an external
    Discord channel. You must have a valid Discord bot application
    ( https://discord.com/developers/applications ) and your DISCORD_BOT_TOKEN
    must be added to settings. (Please put it in secret_settings !)
    """

    key = "discord2chan"
    aliases = ("discord",)
    switch_options = (
        "channel",
        "delete",
        "guild",
        "list",
        "remove",
        "start",
    )
    locks = "cmd:serversetting(DISCORD_ENABLED) and pperm(Developer)"
    help_category = "Comms"

    def func(self):
        """Manage the Evennia<->Discord channel links"""

        if not settings.DISCORD_BOT_TOKEN:
            self.msg(
                "You must add your Discord bot application token to settings as DISCORD_BOT_TOKEN"
            )
            return

        discord_bot = [
            bot for bot in AccountDB.objects.filter(db_is_bot=True, username="DiscordBot")
        ]
        if not discord_bot:
            # create a new discord bot
            bot_class = class_from_module(settings.DISCORD_BOT_CLASS, fallback=bots.DiscordBot)
            discord_bot = create.create_account("DiscordBot", None, None, typeclass=bot_class)
            discord_bot.start()
            self.msg("Created and initialized a new Discord relay bot.")
        else:
            discord_bot = discord_bot[0]

        if not discord_bot.is_typeclass(settings.DISCORD_BOT_CLASS, exact=True):
            self.msg(
                f"WARNING: The Discord bot's typeclass is '{discord_bot.typeclass_path}'. This does"
                f" not match {settings.DISCORD_BOT_CLASS} in settings!"
            )

        if "start" in self.switches:
            if discord_bot.sessions.all():
                self.msg("The Discord bot is already running.")
            else:
                discord_bot.start()
                self.msg("Starting the Discord bot session.")
            return

        if "guild" in self.switches:
            discord_bot.db.tag_guild = not discord_bot.db.tag_guild
            self.msg(
                f"Messages to Evennia |wwill {'' if discord_bot.db.tag_guild else 'not '}|ninclude"
                " the Discord server."
            )
            return
        if "channel" in self.switches:
            discord_bot.db.tag_channel = not discord_bot.db.tag_channel
            self.msg(
                f"Relayed messages |wwill {'' if discord_bot.db.tag_channel else 'not '}|ninclude"
                " the originating channel."
            )
            return

        if "list" in self.switches or not self.args:
            # show all connections
            if channel_list := discord_bot.db.channels:
                table = self.styled_table(
                    "|wLink ID|n",
                    "|wEvennia|n",
                    "|wDiscord|n",
                    border="cells",
                    maxwidth=_DEFAULT_WIDTH,
                )
                # iterate through the channel links
                # load in the pretty names for the discord channels from cache
                dc_chan_names = discord_bot.attributes.get("discord_channels", {})
                for i, (evchan, dcchan) in enumerate(channel_list):
                    dc_info = dc_chan_names.get(dcchan, {"name": dcchan, "guild": "unknown"})
                    table.add_row(
                        i, evchan, f"#{dc_info.get('name','?')}@{dc_info.get('guild','?')}"
                    )
                self.msg(table)
            else:
                self.msg("No Discord connections found.")
            return

        if "disconnect" in self.switches or "remove" in self.switches or "delete" in self.switches:
            if channel_list := discord_bot.db.channels:
                try:
                    lid = int(self.args.strip())
                except ValueError:
                    self.msg("Usage: discord2chan/remove <link id>")
                    return
                if lid < len(channel_list):
                    ev_chan, dc_chan = discord_bot.db.channels.pop(lid)
                    dc_chan_names = discord_bot.attributes.get("discord_channels", {})
                    dc_info = dc_chan_names.get(dc_chan, {"name": "unknown", "guild": "unknown"})
                    self.msg(
                        f"Removed link between {ev_chan} and"
                        f" #{dc_info.get('name','?')}@{dc_info.get('guild','?')}"
                    )
                    return
            else:
                self.msg("There are no active connections to Discord.")
                return

        ev_channel = self.lhs
        dc_channel = self.rhs

        if ev_channel and not dc_channel:
            # show all discord channels linked to self.lhs
            if channel_list := discord_bot.db.channels:
                table = self.styled_table(
                    "|wLink ID|n",
                    "|wEvennia|n",
                    "|wDiscord|n",
                    border="cells",
                    maxwidth=_DEFAULT_WIDTH,
                )
                # iterate through the channel links
                # load in the pretty names for the discord channels from cache
                dc_chan_names = discord_bot.attributes.get("discord_channels", {})
                results = False
                for i, (evchan, dcchan) in enumerate(channel_list):
                    if evchan.lower() == ev_channel.lower():
                        dc_info = dc_chan_names.get(dcchan, {"name": dcchan, "guild": "unknown"})
                        table.add_row(i, evchan, f"#{dc_info['name']}@{dc_info['guild']}")
                        results = True
                if results:
                    self.msg(table)
                else:
                    self.msg(f"There are no Discord channels connected to {ev_channel}.")
            else:
                self.msg("There are no active connections to Discord.")
            return

        # check if link already exists
        if channel_list := discord_bot.db.channels:
            if (ev_channel, dc_channel) in channel_list:
                self.msg("Those channels are already linked.")
                return
        else:
            discord_bot.db.channels = []
        # create the new link
        channel_obj = search.search_channel(ev_channel)
        if not channel_obj:
            self.msg(f"There is no channel '{ev_channel}'")
            return
        channel_obj = channel_obj[0]
        discord_bot.db.channels.append((channel_obj.name, dc_channel))
        channel_obj.connect(discord_bot)
        if dc_chans := discord_bot.db.discord_channels:
            dc_channel_name = dc_chans.get(dc_channel, {}).get("name", dc_channel)
        else:
            dc_channel_name = dc_channel
        self.msg(f"Discord connection created: {channel_obj.name} <-> #{dc_channel_name}.")



"""
This connects to an IRC network/channel and launches an 'bot' onto it.
The bot then pipes what is being said between the IRC channel and one or
more Evennia channels.
"""

import re

from twisted.application import internet
from twisted.internet import protocol, reactor
from twisted.words.protocols import irc

from evennia.server.session import Session
from evennia.utils import ansi, logger, utils

# IRC colors

IRC_BOLD = "\002"
IRC_COLOR = "\003"
IRC_RESET = "\017"
IRC_ITALIC = "\026"
IRC_INVERT = "\x16"
IRC_NORMAL = "99"
IRC_UNDERLINE = "37"

IRC_WHITE = "0"
IRC_BLACK = "1"
IRC_DBLUE = "2"
IRC_DGREEN = "3"
IRC_RED = "4"
IRC_DRED = "5"
IRC_DMAGENTA = "6"
IRC_DYELLOW = "7"
IRC_YELLOW = "8"
IRC_GREEN = "9"
IRC_DCYAN = "10"
IRC_CYAN = "11"
IRC_BLUE = "12"
IRC_MAGENTA = "13"
IRC_DGREY = "14"
IRC_GREY = "15"

# obsolete test:

# test evennia->irc:
# |rred |ggreen |yyellow |bblue |mmagenta |ccyan |wwhite |xdgrey
# |Rdred |Gdgreen |Ydyellow |Bdblue |Mdmagenta |Cdcyan |Wlgrey |Xblack
# |[rredbg |[ggreenbg |[yyellowbg |[bbluebg |[mmagentabg |[ccyanbg |[wlgreybg |[xblackbg

# test irc->evennia
# Use Ctrl+C <num> to produce mIRC colors in e.g. irssi

IRC_COLOR_MAP = dict(
    (
        (r"|n", IRC_COLOR + IRC_NORMAL),  # normal mode
        (r"|H", IRC_RESET),  # un-highlight
        (r"|/", "\n"),  # line break
        (r"|t", "    "),  # tab
        (r"|-", "    "),  # fixed tab
        (r"|_", " "),  # space
        (r"|*", IRC_INVERT),  # invert
        (r"|^", ""),  # blinking text
        (r"|h", IRC_BOLD),  # highlight, use bold instead
        (r"|r", IRC_COLOR + IRC_RED),
        (r"|g", IRC_COLOR + IRC_GREEN),
        (r"|y", IRC_COLOR + IRC_YELLOW),
        (r"|b", IRC_COLOR + IRC_BLUE),
        (r"|m", IRC_COLOR + IRC_MAGENTA),
        (r"|c", IRC_COLOR + IRC_CYAN),
        (r"|w", IRC_COLOR + IRC_WHITE),  # pure white
        (r"|x", IRC_COLOR + IRC_DGREY),  # dark grey
        (r"|R", IRC_COLOR + IRC_DRED),
        (r"|G", IRC_COLOR + IRC_DGREEN),
        (r"|Y", IRC_COLOR + IRC_DYELLOW),
        (r"|B", IRC_COLOR + IRC_DBLUE),
        (r"|M", IRC_COLOR + IRC_DMAGENTA),
        (r"|C", IRC_COLOR + IRC_DCYAN),
        (r"|W", IRC_COLOR + IRC_GREY),  # light grey
        (r"|X", IRC_COLOR + IRC_BLACK),  # pure black
        (r"|[r", IRC_COLOR + IRC_NORMAL + "," + IRC_DRED),
        (r"|[g", IRC_COLOR + IRC_NORMAL + "," + IRC_DGREEN),
        (r"|[y", IRC_COLOR + IRC_NORMAL + "," + IRC_DYELLOW),
        (r"|[b", IRC_COLOR + IRC_NORMAL + "," + IRC_DBLUE),
        (r"|[m", IRC_COLOR + IRC_NORMAL + "," + IRC_DMAGENTA),
        (r"|[c", IRC_COLOR + IRC_NORMAL + "," + IRC_DCYAN),
        (r"|[w", IRC_COLOR + IRC_NORMAL + "," + IRC_GREY),  # light grey background
        (r"|[x", IRC_COLOR + IRC_NORMAL + "," + IRC_BLACK),  # pure black background
    )
)
# ansi->irc
RE_ANSI_COLOR = re.compile(r"|".join([re.escape(key) for key in IRC_COLOR_MAP.keys()]), re.DOTALL)
RE_MXP = re.compile(r"\|lc(.*?)\|lt(.*?)\|le", re.DOTALL)
RE_ANSI_ESCAPES = re.compile(r"(%s)" % "|".join(("{{", "%%", "\\\\")), re.DOTALL)
# irc->ansi
_CLR_LIST = [
    re.escape(val) for val in sorted(IRC_COLOR_MAP.values(), key=len, reverse=True) if val.strip()
]
_CLR_LIST = _CLR_LIST[-2:] + _CLR_LIST[:-2]
RE_IRC_COLOR = re.compile(r"|".join(_CLR_LIST), re.DOTALL)
ANSI_COLOR_MAP = dict((tup[1], tup[0]) for tup in IRC_COLOR_MAP.items() if tup[1].strip())


def parse_ansi_to_irc(string):
    """
    Parse |-type syntax and replace with IRC color markers

    Args:
        string (str): String to parse for ANSI colors.

    Returns:
        parsed_string (str): String with replaced ANSI colors.

    """

    def _sub_to_irc(ansi_match):
        return IRC_COLOR_MAP.get(ansi_match.group(), "")

    in_string = utils.to_str(string)
    parsed_string = []
    parts = RE_ANSI_ESCAPES.split(in_string) + [" "]
    for part, sep in zip(parts[::2], parts[1::2]):
        pstring = RE_ANSI_COLOR.sub(_sub_to_irc, part)
        parsed_string.append("%s%s" % (pstring, sep[0].strip()))
    # strip mxp
    parsed_string = RE_MXP.sub(r"\2", "".join(parsed_string))
    return parsed_string


def parse_irc_to_ansi(string):
    """
    Parse IRC mIRC color syntax and replace with Evennia ANSI color markers

    Args:
        string (str): String to parse for IRC colors.

    Returns:
        parsed_string (str): String with replaced IRC colors.

    """

    def _sub_to_ansi(irc_match):
        return ANSI_COLOR_MAP.get(irc_match.group(), "")

    in_string = utils.to_str(string)
    pstring = RE_IRC_COLOR.sub(_sub_to_ansi, in_string)
    return pstring


# IRC bot


class IRCBot(irc.IRCClient, Session):
    """
    An IRC bot that tracks activity in a channel as well
    as sends text to it when prompted

    """

    lineRate = 1

    # assigned by factory at creation

    nickname = None
    logger = None
    factory = None
    channel = None
    sourceURL = "http://code.evennia.com"

    def signedOn(self):
        """
        This is called when we successfully connect to the network. We
        make sure to now register with the game as a full session.

        """
        self.join(self.channel)
        self.stopping = False
        self.factory.bot = self
        address = "%s@%s" % (self.channel, self.network)
        self.init_session("ircbot", address, self.factory.sessionhandler)
        # we link back to our bot and log in
        self.uid = int(self.factory.uid)
        self.logged_in = True
        self.factory.sessionhandler.connect(self)
        logger.log_info(
            "IRC bot '%s' connected to %s at %s:%s."
            % (self.nickname, self.channel, self.network, self.port)
        )

    def disconnect(self, reason=""):
        """
        Called by sessionhandler to disconnect this protocol.

        Args:
            reason (str): Motivation for the disconnect.

        """
        self.sessionhandler.disconnect(self)
        self.stopping = True
        self.transport.loseConnection()

    def at_login(self):
        pass

    def privmsg(self, user, channel, msg):
        """
        Called when the connected channel receives a message.

        Args:
            user (str): User name sending the message.
            channel (str): Channel name seeing the message.
            msg (str): The message arriving from channel.

        """
        if channel == self.nickname:
            # private message
            user = user.split("!", 1)[0]
            self.data_in(text=msg, type="privmsg", user=user, channel=channel)
        elif not msg.startswith("***"):
            # channel message
            user = user.split("!", 1)[0]
            user = ansi.raw(user)
            self.data_in(text=msg, type="msg", user=user, channel=channel)

    def action(self, user, channel, msg):
        """
        Called when an action is detected in channel.

        Args:
            user (str): User name sending the message.
            channel (str): Channel name seeing the message.
            msg (str): The message arriving from channel.

        """
        if not msg.startswith("**"):
            user = user.split("!", 1)[0]
            self.data_in(text=msg, type="action", user=user, channel=channel)

    def get_nicklist(self):
        """
        Retrieve name list from the channel. The return
        is handled by the catch methods below.

        """
        if not self.nicklist:
            self.sendLine("NAMES %s" % self.channel)

    def irc_RPL_NAMREPLY(self, prefix, params):
        """ "Handles IRC NAME request returns (nicklist)"""
        channel = params[2].lower()
        if channel != self.channel.lower():
            return
        self.nicklist += params[3].split(" ")

    def irc_RPL_ENDOFNAMES(self, prefix, params):
        """Called when the nicklist has finished being returned."""
        channel = params[1].lower()
        if channel != self.channel.lower():
            return
        self.data_in(
            text="", type="nicklist", user="server", channel=channel, nicklist=self.nicklist
        )
        self.nicklist = []

    def pong(self, user, time):
        """
        Called with the return timing from a PING.

        Args:
            user (str): Name of user
            time (float): Ping time in secs.

        """
        self.data_in(text="", type="ping", user="server", channel=self.channel, timing=time)

    def data_in(self, text=None, **kwargs):
        """
        Data IRC -> Server.

        Keyword Args:
            text (str): Ingoing text.
            kwargs (any): Other data from protocol.

        """
        self.sessionhandler.data_in(self, bot_data_in=[parse_irc_to_ansi(text), kwargs])

    def send_channel(self, *args, **kwargs):
        """
        Send channel text to IRC channel (visible to all). Note that
        we don't handle the "text" send (it's rerouted to send_default
        which does nothing) - this is because the IRC bot is a normal
        session and would otherwise report anything that happens to it
        to the IRC channel (such as it seeing server reload messages).

        Args:
            text (str): Outgoing text

        """
        text = args[0] if args else ""
        if text:
            text = parse_ansi_to_irc(text)
            self.say(self.channel, text)

    def send_privmsg(self, *args, **kwargs):
        """
        Send message only to specific user.

        Args:
            text (str): Outgoing text.

        Keyword Args:
            user (str): the nick to send
                privately to.

        """
        text = args[0] if args else ""
        user = kwargs.get("user", None)
        if text and user:
            text = parse_ansi_to_irc(text)
            self.msg(user, text)

    def send_request_nicklist(self, *args, **kwargs):
        """
        Send a request for the channel nicklist. The return (handled
        by `self.irc_RPL_ENDOFNAMES`) will be sent back as a message
        with type `nicklist'.
        """
        self.get_nicklist()

    def send_ping(self, *args, **kwargs):
        """
        Send a ping. The return (handled by `self.pong`) will be sent
        back as a message of type 'ping'.
        """
        self.ping(self.nickname)

    def send_reconnect(self, *args, **kwargs):
        """
        The server instructs us to rebuild the connection by force,
        probably because the client silently lost connection.
        """
        self.factory.reconnect()

    def send_default(self, *args, **kwargs):
        """
        Ignore other types of sends.

        """
        pass


class IRCBotFactory(protocol.ReconnectingClientFactory):
    """
    Creates instances of IRCBot, connecting with a staggered
    increase in delay

    """

    # scaling reconnect time
    initialDelay = 1
    factor = 1.5
    maxDelay = 60

    def __init__(
        self,
        sessionhandler,
        uid=None,
        botname=None,
        channel=None,
        network=None,
        port=None,
        ssl=None,
    ):
        """
        Storing some important protocol properties.

        Args:
            sessionhandler (SessionHandler): Reference to the main Sessionhandler.

        Keyword Args:
            uid (int): Bot user id.
            botname (str): Bot name (seen in IRC channel).
            channel (str): IRC channel to connect to.
            network (str): Network address to connect to.
            port (str): Port of the network.
            ssl (bool): Indicates SSL connection.

        """
        self.sessionhandler = sessionhandler
        self.uid = uid
        self.nickname = str(botname)
        self.channel = str(channel)
        self.network = str(network)
        self.port = port
        self.ssl = ssl
        self.bot = None
        self.nicklists = {}

    def buildProtocol(self, addr):
        """
        Build the protocol and assign it some properties.

        Args:
            addr (str): Not used; using factory data.

        """
        protocol = IRCBot()
        protocol.factory = self
        protocol.nickname = self.nickname
        protocol.channel = self.channel
        protocol.network = self.network
        protocol.port = self.port
        protocol.ssl = self.ssl
        protocol.nicklist = []
        return protocol

    def startedConnecting(self, connector):
        """
        Tracks reconnections for debugging.

        Args:
            connector (Connector): Represents the connection.

        """
        logger.log_info("(re)connecting to %s" % self.channel)

    def clientConnectionFailed(self, connector, reason):
        """
        Called when Client failed to connect.

        Args:
            connector (Connection): Represents the connection.
            reason (str): The reason for the failure.

        """
        self.retry(connector)

    def clientConnectionLost(self, connector, reason):
        """
        Called when Client loses connection.

        Args:
            connector (Connection): Represents the connection.
            reason (str): The reason for the failure.

        """
        if not (self.bot or (self.bot and self.bot.stopping)):
            self.retry(connector)

    def reconnect(self):
        """
        Force a reconnection of the bot protocol. This requires
        de-registering the session and then reattaching a new one,
        otherwise you end up with an ever growing number of bot
        sessions.

        """
        self.bot.stopping = True
        self.bot.transport.loseConnection()
        self.sessionhandler.server_disconnect(self.bot)
        self.start()

    def start(self):
        """
        Connect session to sessionhandler.

        """
        if self.port:
            if self.ssl:
                try:
                    from twisted.internet import ssl

                    service = reactor.connectSSL(
                        self.network, int(self.port), self, ssl.ClientContextFactory()
                    )
                except ImportError:
                    logger.log_err("To use SSL, the PyOpenSSL module must be installed.")
            else:
                service = internet.TCPClient(self.network, int(self.port), self)
            self.sessionhandler.portal.services.addService(service)
